# 25 "src/kernel_internals/parsing/logic_preprocess.mll"
 
  open Lexing
  type end_of_buffer = NEWLINE | SPACE | CHAR
  let preprocess_buffer = Buffer.create 1024
  let output_buffer = Buffer.create 1024
  let beg_of_line = Buffer.create 8
  (* Standard prohibits the predefined macros to be subject of a #define
     (or #undef) directive. We thus have to filter the definition of these
     macros from gcc's output (gcc emits a warning otherwise).
     The list of predefined macros is taken from C11 standard, in the order
     in which they are defined in Section 6.10.8
   *)
  let blacklisted_macros = [
    (* 6.10.8.1 mandatory macros. *)
    "__DATE__"; "__FILE"; "__LINE__"; "__STDC__"; "__STDC_HOSTED__";
    "__STDC_VERSION__"; "__TIME__";
    (* 6.10.8.2 environment macros *)
    "__STDC_ISO_10646__"; "__STDC_MB_MIGHT_NEQ_WC__";
    "__STDC_UTF_16__"; "__STDC_UTF_32__";
    (* 6.10.8.3 conditional feature macros *)
    "__STDC_ANALYZABLE__"; "__STDC_IEC_559__"; "__STDC_IEC_559_COMPLEX__";
    "__STDC_LIB_EXT1__"; "__STD_NO_ATOMICS__"; "__STD_NO_COMPLEX__";
    "__STDC_NO_THREADS__"; "__STDC_NO_VLA__";
    (* expanding assert, an ACSL keyword, is not a good idea. *)
    "assert"]
  let is_newline = ref CHAR
  let curr_file = ref ""
  let curr_line = ref 1
  let has_annot = ref false

  let reset () =
    Buffer.clear preprocess_buffer;
    Buffer.clear output_buffer;
    Buffer.clear beg_of_line;
    is_newline := CHAR;
    curr_file := "";
    curr_line := 1;
    has_annot := false

  let backslash = "__ANNOT_BACKSLASH__"
  let annot_content = "__ANNOT_CONTENT__"

  let re_backslash = Str.regexp_string backslash
  let re_annot_content = Str.regexp_string annot_content

  (* Delimiters for the various annotations in the preprocessing buffer.
     We have one delimiter for the beginning of an annotation (to discard
     #defines along the way), and three delimiters for the various ways
     an annotation can end:
      - on a normal line
      - with a newline
      - with a newline inside a comment (only for one-line annotations)
     When preprocessed annotations are inserted back in the main file, this will
     result in distinct translation to preserve line numbers while avoiding
     ill-formed annotations.
  *)
  let annot_beg =         "////////////////__ANNOT_BEG__"
  let annot_end =         "////////////////__ANNOT_END__"
  let annot_end_nl  =     "////////////////__ANNOT_END_NL__"
  let annot_end_comment = "////////////////__ANNOT_END_COMMENT__"

  let abort_preprocess reason =
    let source = {Lexing.dummy_pos with Lexing.pos_fname = !curr_file;
                  pos_lnum = !curr_line;}
    in
    Kernel.error ~source
      "Can't preprocess annotation: %s\nSome annotations will be kept as is"
      reason

  let next_preprocessed file =
    let content = Buffer.create 80 in
    let rec ignore_content () =
      let s = input_line file in
      if s <> annot_beg then ignore_content ()
    in
    let rec get_annot first =
      let s = input_line file in
      if s = annot_end then false, Buffer.contents content
      else if s = annot_end_nl then true, Buffer.contents content
      else if s = annot_end_comment then begin
        Buffer.add_char content '\n';
        false, Buffer.contents content
      end else begin
        if not first then Buffer.add_char content '\n';
        Buffer.add_string content s;
        get_annot false
      end
    in
    let replace_backslash s = Str.global_replace re_backslash "\\\\" s in
    try
      ignore_content ();
      ignore (input_line file); (* ignore the #line directive *)
      let with_nl, content = get_annot true in
      with_nl, replace_backslash content
    with End_of_file ->
      Kernel.fatal
        "too few annotations in result file while pre-processing annotations"

  let output_result outfile preprocessed content =
    let rec aux = function
      | [] -> ()
      | [s] -> output_string outfile s
      | content :: rem ->
          output_string outfile content;
          output_string outfile "/*@";
          let with_nl, pp_content = next_preprocessed preprocessed in
          output_string outfile pp_content;
          output_string outfile "*/";
          if with_nl then output_char outfile '\n';
          aux rem
    in aux content

  let preprocess_annots suffix cpp outfile =
    if !has_annot then begin
      let debug =
        Kernel.debug_atleast 3 ||
          Kernel.Debug_category.exists (fun x -> x = "parser")
      in
      let ppname =
        try Extlib.temp_file_cleanup_at_exit ~debug "ppannot" suffix
        with Extlib.Temp_file_error s ->
          Kernel.abort
            "Could not open temporary file for logic pre-processing: %s" s
      in
      let ppfile = open_out ppname in
      Buffer.output_buffer ppfile preprocess_buffer;
      close_out ppfile;
      let cppname = Extlib.temp_file_cleanup_at_exit ~debug "cppannot" suffix in
      let res = Sys.command (cpp ppname cppname) in
      let result_file =
        if res <> 0 then begin
          abort_preprocess "Preprocessor call exited with an error";
          if not debug then Extlib.safe_remove cppname;
          ppname
        end else cppname
      in
      let result = open_in result_file in
      let content =
        Str.split_delim re_annot_content (Buffer.contents output_buffer)
      in
      output_result outfile result content;
    end else begin
      Buffer.output_buffer outfile output_buffer
    end;
    flush outfile

  let add_preprocess_line_info () =
    Printf.bprintf
      preprocess_buffer "# %d %s \n%s   "
      !curr_line !curr_file (Buffer.contents beg_of_line);
    Buffer.clear beg_of_line

  let make_newline () =
    incr curr_line;
    Buffer.clear beg_of_line

# 159 "src/kernel_internals/parsing/logic_preprocess.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\245\255\246\255\247\255\248\255\249\255\001\000\004\000\
    \017\000\000\000\031\000\000\000\000\000\000\000\001\000\001\000\
    \002\000\081\000\160\000\000\000\006\000\006\000\004\000\008\000\
    \018\001\028\000\010\000\030\000\031\000\254\255\029\000\032\000\
    \252\255\000\000\253\255\250\255\001\000\251\255\081\000\250\255\
    \251\255\252\255\253\255\003\000\001\000\255\255\254\255\036\000\
    \253\255\003\000\255\255\254\255\082\000\250\255\251\255\252\255\
    \253\255\004\000\254\255\255\255\212\000\250\255\251\255\252\255\
    \253\255\079\000\254\255\255\255\085\000\251\255\253\255\254\255\
    \083\000\252\255\255\255\213\000\251\255\253\255\254\255\215\000\
    \252\255\255\255\044\001\247\255\248\255\249\255\250\255\251\255\
    \252\255\028\000\254\255\029\000\255\255\253\255\130\000\252\255\
    \253\255\030\000\255\255\254\255\023\001\250\255\251\255\251\000\
    \254\255\255\255\252\255\253\255\025\001\251\255\252\255\003\000\
    \254\255\255\255\253\255\030\001\252\255\253\255\254\255\255\255\
    \045\001\250\255\047\000\252\255\253\255\254\255\255\255\251\255\
    \088\000\254\255\255\255\089\000\254\255\255\255";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\010\000\010\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\005\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\005\000\005\000\255\255\255\255\255\255\
    \255\255\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\005\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\005\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \004\000\255\255\255\255\255\255\255\255\255\255\255\255\004\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\008\000\255\255\008\000\255\255\255\255\255\255\255\255\
    \255\255\003\000\255\255\255\255\255\255\255\255\255\255\005\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\004\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\005\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\000\000\000\000\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\027\000\255\255\027\000\028\000\000\000\035\000\032\000\
    \000\000\255\255\000\000\000\000\255\255\000\000\039\000\000\000\
    \000\000\000\000\000\000\255\255\255\255\000\000\000\000\048\000\
    \000\000\255\255\000\000\000\000\053\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\061\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\069\000\000\000\000\000\000\000\
    \255\255\000\000\000\000\076\000\000\000\000\000\000\000\255\255\
    \000\000\000\000\083\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\255\255\000\000\000\000\095\000\000\000\
    \000\000\255\255\000\000\000\000\101\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\000\000\109\000\000\000\000\000\255\255\
    \000\000\000\000\000\000\116\000\000\000\000\000\000\000\000\000\
    \121\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \129\000\000\000\000\000\132\000\000\000\000\000";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\005\000\045\000\000\000\010\000\058\000\000\000\
    \000\000\000\000\000\000\026\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\026\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\007\000\010\000\114\000\059\000\002\000\
    \010\000\029\000\026\000\031\000\025\000\046\000\050\000\006\000\
    \030\000\026\000\051\000\025\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\255\255\010\000\
    \028\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\093\000\092\000\099\000\049\000\008\000\
    \008\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\066\000\017\000\040\000\056\000\036\000\127\000\070\000\
    \033\000\000\000\130\000\133\000\020\000\013\000\014\000\017\000\
    \012\000\022\000\015\000\021\000\017\000\024\000\019\000\016\000\
    \009\000\017\000\023\000\042\000\054\000\074\000\067\000\071\000\
    \041\000\011\000\000\000\034\000\037\000\034\000\037\000\000\000\
    \043\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\009\000\098\000\000\000\000\000\000\000\
    \000\000\000\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\097\000\044\000\057\000\073\000\
    \018\000\072\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\000\000\000\000\000\000\000\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\000\000\000\000\000\000\000\000\064\000\077\000\
    \000\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\062\000\078\000\000\000\081\000\018\000\
    \004\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\024\000\255\255\255\255\255\255\255\255\
    \255\255\105\000\107\000\113\000\255\255\000\000\000\000\000\000\
    \118\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \065\000\079\000\024\000\080\000\000\000\000\000\090\000\126\000\
    \000\000\000\000\000\000\112\000\000\000\000\000\104\000\000\000\
    \000\000\000\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\087\000\119\000\084\000\123\000\
    \000\000\255\255\055\000\085\000\124\000\255\255\091\000\106\000\
    \130\000\133\000\000\000\089\000\122\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\088\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\103\000\000\000\111\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\096\000\000\000\000\000\000\000\000\000\000\000\
    \086\000\125\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\063\000\255\255\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\102\000\
    \000\000\110\000\000\000\000\000\000\000\000\000\117\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\126\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\044\000\255\255\007\000\057\000\255\255\
    \255\255\255\255\255\255\026\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\000\000\007\000\111\000\057\000\000\000\
    \010\000\028\000\026\000\006\000\026\000\043\000\047\000\000\000\
    \006\000\008\000\049\000\008\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\007\000\007\000\007\000\007\000\025\000\010\000\
    \027\000\008\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\089\000\091\000\097\000\047\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\010\000\010\000\
    \010\000\065\000\017\000\038\000\052\000\030\000\122\000\068\000\
    \031\000\255\255\128\000\131\000\019\000\012\000\013\000\016\000\
    \007\000\009\000\014\000\020\000\021\000\023\000\011\000\015\000\
    \007\000\017\000\022\000\038\000\052\000\072\000\065\000\068\000\
    \038\000\007\000\255\255\033\000\036\000\033\000\036\000\255\255\
    \038\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\010\000\094\000\255\255\255\255\255\255\
    \255\255\255\255\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\094\000\038\000\052\000\072\000\
    \017\000\068\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\017\000\017\000\017\000\017\000\
    \017\000\017\000\017\000\017\000\255\255\255\255\255\255\255\255\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\255\255\255\255\255\255\255\255\060\000\075\000\
    \255\255\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\060\000\075\000\255\255\079\000\018\000\
    \000\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\018\000\018\000\018\000\018\000\018\000\
    \018\000\018\000\018\000\024\000\025\000\030\000\027\000\028\000\
    \031\000\100\000\103\000\108\000\047\000\255\255\255\255\255\255\
    \115\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \060\000\075\000\024\000\079\000\255\255\255\255\082\000\120\000\
    \255\255\255\255\255\255\108\000\255\255\255\255\100\000\255\255\
    \255\255\255\255\024\000\024\000\024\000\024\000\024\000\024\000\
    \024\000\024\000\024\000\024\000\082\000\115\000\082\000\120\000\
    \255\255\038\000\052\000\082\000\120\000\068\000\082\000\103\000\
    \128\000\131\000\255\255\082\000\120\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\082\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\100\000\255\255\108\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\094\000\255\255\255\255\255\255\255\255\255\255\
    \082\000\120\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\060\000\075\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\100\000\
    \255\255\108\000\255\255\255\255\255\255\255\255\115\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\082\000\120\000";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \010\000\000\000\036\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\000\000\000\000\000\000\000\000\002\000\
    \062\000\000\000\002\000\002\000\000\000\026\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\020\000\020\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\017\000\009\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\009\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \001\000\017\000\009\000\001\000\023\000\000\000\000\000\000\000\
    \000\000\000\000\009\000\000\000\000\000\001\000\000\000\000\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\001\000\000\000\000\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\012\000\012\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\001\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\017\000\017\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\012\000\012\000\
    \012\000\012\000\012\000\012\000\012\000\012\000\012\000\012\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\007\000\017\000\026\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\008\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \007\000\017\000\026\000\000\000\027\000\255\255\255\255\255\255\
    \255\255\255\255\008\000\255\255\255\255\010\000\255\255\255\255\
    \007\000\007\000\007\000\007\000\007\000\007\000\007\000\007\000\
    \007\000\007\000\008\000\008\000\008\000\008\000\008\000\008\000\
    \008\000\008\000\008\000\008\000\010\000\255\255\255\255\024\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\010\000\010\000\010\000\010\000\
    \010\000\010\000\010\000\010\000\010\000\010\000\024\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\016\000\021\000\023\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\024\000\024\000\
    \024\000\024\000\024\000\024\000\024\000\024\000\024\000\024\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\004\255\255\006\255\005\255\255\006\255\255\005\255\006\255\
    \255\007\255\255\000\007\255\008\255\255\000\004\001\005\002\006\
    \003\008\255";
}

let rec main lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 9 (-1) ;   __ocaml_lex_main_rec lexbuf 0
and __ocaml_lex_main_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 183 "src/kernel_internals/parsing/logic_preprocess.mll"
                                                                      m
# 507 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 184 "src/kernel_internals/parsing/logic_preprocess.mll"
      (
        let blacklisted = List.mem m blacklisted_macros in
        if not blacklisted then
          Buffer.add_string preprocess_buffer (lexeme lexbuf);
        macro blacklisted lexbuf
      )
# 516 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
let
# 190 "src/kernel_internals/parsing/logic_preprocess.mll"
                                                       line
# 522 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_mem.(1)
and
# 191 "src/kernel_internals/parsing/logic_preprocess.mll"
                                     file
# 527 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(2) lexbuf.Lexing.lex_mem.(3) in
# 192 "src/kernel_internals/parsing/logic_preprocess.mll"
    ( (try
        curr_line := (int_of_string line) -1
       with Failure "int_of_string" -> curr_line:= -1);
      if file <> "" then curr_file := file;
      Buffer.add_string output_buffer (lexeme lexbuf);
      make_newline();
      main lexbuf
    )
# 538 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
let
# 200 "src/kernel_internals/parsing/logic_preprocess.mll"
                        c
# 544 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 3) in
# 200 "src/kernel_internals/parsing/logic_preprocess.mll"
                           ( (* Skip special doxygen comments. Use of '@'
                                instead of !Clexer.annot_char is intentional *)
        Buffer.add_string beg_of_line "   ";
        Buffer.add_string output_buffer (lexeme lexbuf);
        comment c lexbuf;)
# 552 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
let
# 205 "src/kernel_internals/parsing/logic_preprocess.mll"
                c
# 558 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 205 "src/kernel_internals/parsing/logic_preprocess.mll"
                   (
      if c = !Clexer.annot_char then begin
        is_newline:=CHAR;
        has_annot := true;
        Buffer.add_string output_buffer annot_content;
        Buffer.add_string preprocess_buffer annot_beg;
        Buffer.add_char preprocess_buffer '\n';
        add_preprocess_line_info();
        annot lexbuf
      end else begin
        if c = '\n' then make_newline()
        else Buffer.add_string beg_of_line "   ";
        Buffer.add_string output_buffer (lexeme lexbuf);
        comment c lexbuf;
      end)
# 576 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 220 "src/kernel_internals/parsing/logic_preprocess.mll"
                      ( (* See comments for "/*@{" above *)
        Buffer.add_string output_buffer (lexeme lexbuf);
        oneline_comment lexbuf;
      )
# 584 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 224 "src/kernel_internals/parsing/logic_preprocess.mll"
                c
# 590 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 224 "src/kernel_internals/parsing/logic_preprocess.mll"
                   (
      if c = !Clexer.annot_char then begin
        is_newline:=CHAR;
        has_annot:=true;
        Buffer.add_string output_buffer annot_content;
        Buffer.add_string preprocess_buffer annot_beg;
        Buffer.add_char preprocess_buffer '\n';
        add_preprocess_line_info();
        oneline_annot lexbuf
      end
      else if c = '\n' then begin
        make_newline ();
        Buffer.add_string output_buffer (lexeme lexbuf);
        main lexbuf
      end
      else begin
        Buffer.add_string output_buffer (lexeme lexbuf);
        oneline_comment lexbuf;
      end)
# 612 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 6 ->
# 243 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
      make_newline (); Buffer.add_char output_buffer '\n'; main lexbuf )
# 618 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 7 ->
# 245 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( )
# 623 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 8 ->
# 246 "src/kernel_internals/parsing/logic_preprocess.mll"
        (
      Buffer.add_char beg_of_line ' ';
      Buffer.add_char output_buffer '"'; 
      c_string lexbuf )
# 631 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 9 ->
# 250 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( 
      Buffer.add_char beg_of_line ' ';
      Buffer.add_char output_buffer '\'';
      c_char lexbuf )
# 639 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 10 ->
let
# 254 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 645 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 254 "src/kernel_internals/parsing/logic_preprocess.mll"
           (
      Buffer.add_char beg_of_line ' ';
      Buffer.add_char output_buffer c;
      main lexbuf )
# 652 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_main_rec lexbuf __ocaml_lex_state

and macro blacklisted lexbuf =
    __ocaml_lex_macro_rec blacklisted lexbuf 38
and __ocaml_lex_macro_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 259 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
      make_newline ();
      Buffer.add_char output_buffer '\n';
      macro blacklisted lexbuf
    )
# 668 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 266 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( macro_comment blacklisted lexbuf )
# 673 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 267 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( 
  if not blacklisted then
    Buffer.add_char preprocess_buffer '"';
  macro_string blacklisted lexbuf
)
# 682 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 272 "src/kernel_internals/parsing/logic_preprocess.mll"
      (
  if not blacklisted then
    Buffer.add_char preprocess_buffer '\'';
  macro_char blacklisted lexbuf
)
# 691 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 277 "src/kernel_internals/parsing/logic_preprocess.mll"
       (
      if not blacklisted then
        Buffer.add_char preprocess_buffer '\n';
      make_newline ();
      Buffer.add_char output_buffer '\n';
      main lexbuf
    )
# 702 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 284 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 708 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 284 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
           if not blacklisted then
             Buffer.add_char preprocess_buffer c;
           macro blacklisted lexbuf
         )
# 716 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_macro_rec blacklisted lexbuf __ocaml_lex_state

and macro_comment blacklisted lexbuf =
    __ocaml_lex_macro_comment_rec blacklisted lexbuf 47
and __ocaml_lex_macro_comment_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 290 "src/kernel_internals/parsing/logic_preprocess.mll"
       (
      make_newline ();

      macro_comment blacklisted lexbuf
    )
# 732 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 295 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( macro blacklisted lexbuf )
# 737 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 296 "src/kernel_internals/parsing/logic_preprocess.mll"
     ( macro_comment blacklisted lexbuf )
# 742 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_macro_comment_rec blacklisted lexbuf __ocaml_lex_state

and macro_string blacklisted lexbuf =
    __ocaml_lex_macro_string_rec blacklisted lexbuf 52
and __ocaml_lex_macro_string_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 299 "src/kernel_internals/parsing/logic_preprocess.mll"
             s
# 755 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 299 "src/kernel_internals/parsing/logic_preprocess.mll"
               (
  if not blacklisted then Buffer.add_string preprocess_buffer s;
  macro_string blacklisted lexbuf
  )
# 762 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 303 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
  make_newline();
  Buffer.add_char output_buffer '\n';
  macro_string blacklisted lexbuf
)
# 771 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 308 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( abort_preprocess "unterminated string in macro definition" )
# 776 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 309 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( abort_preprocess "unterminated string in macro definition" )
# 781 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 310 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( if not blacklisted then Buffer.add_char preprocess_buffer '"';
        macro blacklisted lexbuf )
# 787 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 312 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 793 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 312 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( if not blacklisted then Buffer.add_char preprocess_buffer c;
           macro_string blacklisted lexbuf )
# 798 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_macro_string_rec blacklisted lexbuf __ocaml_lex_state

and macro_char blacklisted lexbuf =
    __ocaml_lex_macro_char_rec blacklisted lexbuf 60
and __ocaml_lex_macro_char_rec blacklisted lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
let
# 315 "src/kernel_internals/parsing/logic_preprocess.mll"
            s
# 811 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_start_pos (lexbuf.Lexing.lex_start_pos + 2) in
# 315 "src/kernel_internals/parsing/logic_preprocess.mll"
              (
  if not blacklisted then Buffer.add_string preprocess_buffer s;
  macro_char blacklisted lexbuf
  )
# 818 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 319 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
  make_newline();
  Buffer.add_char output_buffer '\n';
  macro_char blacklisted lexbuf
)
# 827 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 324 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( abort_preprocess "unterminated char in macro definition" )
# 832 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 325 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( abort_preprocess "unterminated char in macro definition" )
# 837 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 326 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( if not blacklisted then Buffer.add_char preprocess_buffer '\'';
        macro blacklisted lexbuf )
# 843 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 328 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 849 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 328 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( if not blacklisted then Buffer.add_char preprocess_buffer c;
           macro_char blacklisted lexbuf )
# 854 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_macro_char_rec blacklisted lexbuf __ocaml_lex_state

and c_string lexbuf =
    __ocaml_lex_c_string_rec lexbuf 68
and __ocaml_lex_c_string_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 331 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           Buffer.add_string output_buffer (lexeme lexbuf);
           c_string lexbuf )
# 868 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 334 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( Buffer.add_char beg_of_line ' '; 
         Buffer.add_char output_buffer '"';
         main lexbuf )
# 875 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 337 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( make_newline ();
         Buffer.add_char output_buffer '\n';
         c_string lexbuf
       )
# 883 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 341 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           Buffer.add_string output_buffer (lexeme lexbuf);
           c_string lexbuf )
# 890 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
let
# 344 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 896 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 344 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char beg_of_line ' ';
           Buffer.add_char output_buffer c;
           c_string lexbuf )
# 902 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_c_string_rec lexbuf __ocaml_lex_state

and c_char lexbuf =
    __ocaml_lex_c_char_rec lexbuf 75
and __ocaml_lex_c_char_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 349 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           Buffer.add_string output_buffer (lexeme lexbuf);
           c_char lexbuf )
# 916 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 352 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( Buffer.add_char beg_of_line ' '; 
         Buffer.add_char output_buffer '\'';
         main lexbuf )
# 923 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 355 "src/kernel_internals/parsing/logic_preprocess.mll"
       ( make_newline ();
         Buffer.add_char output_buffer '\n';
         c_char lexbuf
       )
# 931 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 359 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string beg_of_line "  ";
           Buffer.add_string output_buffer (lexeme lexbuf);
           c_char lexbuf )
# 938 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
let
# 362 "src/kernel_internals/parsing/logic_preprocess.mll"
       c
# 944 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 362 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char beg_of_line ' ';
           Buffer.add_char output_buffer c;
           c_char lexbuf )
# 950 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_c_char_rec lexbuf __ocaml_lex_state

and annot lexbuf =
    __ocaml_lex_annot_rec lexbuf 82
and __ocaml_lex_annot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 367 "src/kernel_internals/parsing/logic_preprocess.mll"
          (
      if !is_newline = NEWLINE then
        Buffer.add_string preprocess_buffer annot_end_nl
      else begin
        Buffer.add_char preprocess_buffer '\n';
        Buffer.add_string preprocess_buffer annot_end;
      end;
      Buffer.add_char preprocess_buffer '\n';
      main lexbuf )
# 970 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 376 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( is_newline := NEWLINE;
           incr curr_line;
           Buffer.add_char preprocess_buffer '\n';
           annot lexbuf )
# 978 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 380 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string preprocess_buffer "//";
           annot_comment lexbuf )
# 984 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 382 "src/kernel_internals/parsing/logic_preprocess.mll"
        (
      if !is_newline = NEWLINE then is_newline:=SPACE;
      Buffer.add_char preprocess_buffer ' ';
      annot lexbuf )
# 992 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 386 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
      if !is_newline = NEWLINE then is_newline:=SPACE;
      Buffer.add_char preprocess_buffer ' ';
      annot lexbuf )
# 1000 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
# 394 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( 
        is_newline := CHAR;
        Buffer.add_string preprocess_buffer backslash;
        annot lexbuf )
# 1008 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 6 ->
# 398 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
        is_newline := CHAR;
        Buffer.add_char preprocess_buffer '\'';
        char annot lexbuf )
# 1016 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 7 ->
# 402 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
        is_newline:=CHAR;
        Buffer.add_char preprocess_buffer '"';
        string annot lexbuf )
# 1024 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 8 ->
let
# 406 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1030 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 406 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline := CHAR;
             Buffer.add_char preprocess_buffer c;
             annot lexbuf )
# 1036 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_rec lexbuf __ocaml_lex_state

and annot_comment lexbuf =
    __ocaml_lex_annot_comment_rec lexbuf 94
and __ocaml_lex_annot_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 411 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char preprocess_buffer '\n';
           annot lexbuf
         )
# 1051 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 415 "src/kernel_internals/parsing/logic_preprocess.mll"
         (
        Buffer.add_char preprocess_buffer '\n';
        Buffer.add_string preprocess_buffer annot_end;
        Buffer.add_char preprocess_buffer '\n';
        main lexbuf )
# 1060 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 420 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof in the middle of a comment" )
# 1065 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
let
# 421 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1071 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 421 "src/kernel_internals/parsing/logic_preprocess.mll"
           (
    Buffer.add_char preprocess_buffer c; annot_comment lexbuf )
# 1076 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_comment_rec lexbuf __ocaml_lex_state

and char annot lexbuf =
    __ocaml_lex_char_rec annot lexbuf 100
and __ocaml_lex_char_rec annot lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 426 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char preprocess_buffer '\n';
           char annot lexbuf
         )
# 1091 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 430 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( is_newline:=CHAR;
           Buffer.add_char preprocess_buffer '\'';
           annot lexbuf )
# 1098 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 433 "src/kernel_internals/parsing/logic_preprocess.mll"
          ( is_newline:=CHAR;
            Buffer.add_string preprocess_buffer "\\'";
            char annot lexbuf )
# 1105 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 436 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
            Buffer.add_string preprocess_buffer "\\\\";
            char annot lexbuf )
# 1112 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 439 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing a char literal" )
# 1117 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 440 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1123 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 440 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_char preprocess_buffer c;
             char annot lexbuf )
# 1129 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_char_rec annot lexbuf __ocaml_lex_state

and string annot lexbuf =
    __ocaml_lex_string_rec annot lexbuf 108
and __ocaml_lex_string_rec annot lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 445 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( incr curr_line; is_newline:=NEWLINE;
           Buffer.add_char preprocess_buffer '\n'; string annot lexbuf
         )
# 1143 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 448 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( is_newline:=CHAR;
          Buffer.add_char preprocess_buffer '"'; annot lexbuf )
# 1149 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 450 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_string preprocess_buffer "\\\"";
             string annot lexbuf )
# 1156 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 453 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing a string literal" )
# 1161 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
let
# 454 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1167 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 454 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( is_newline:=CHAR;
             Buffer.add_char preprocess_buffer c;
             string annot lexbuf )
# 1173 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_string_rec annot lexbuf __ocaml_lex_state

and comment c lexbuf =
    __ocaml_lex_comment_rec c lexbuf 115
and __ocaml_lex_comment_rec c lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 460 "src/kernel_internals/parsing/logic_preprocess.mll"
        (
      Buffer.add_char beg_of_line ' ';
      Buffer.add_char output_buffer  '/';
      if c = '*' then
        main lexbuf
      else
        comment '/' lexbuf
      )
# 1192 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 468 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( make_newline (); Buffer.add_char output_buffer '\n';
           comment '\n' lexbuf )
# 1198 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 470 "src/kernel_internals/parsing/logic_preprocess.mll"
        ( abort_preprocess "eof while parsing C comment" )
# 1203 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
let
# 471 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1209 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 471 "src/kernel_internals/parsing/logic_preprocess.mll"
           (
      Buffer.add_char beg_of_line ' ';
      Buffer.add_char output_buffer c;
      comment c lexbuf)
# 1216 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec c lexbuf __ocaml_lex_state

and oneline_annot lexbuf =
    __ocaml_lex_oneline_annot_rec lexbuf 120
and __ocaml_lex_oneline_annot_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 477 "src/kernel_internals/parsing/logic_preprocess.mll"
             (
      incr curr_line;
      Buffer.add_char preprocess_buffer '\n';
      Buffer.add_string preprocess_buffer annot_end_nl;
      Buffer.add_char preprocess_buffer '\n';
      main lexbuf )
# 1233 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
# 483 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string preprocess_buffer backslash;
           oneline_annot lexbuf )
# 1239 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 2 ->
# 485 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char preprocess_buffer '\'';
           char oneline_annot lexbuf )
# 1245 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 3 ->
# 487 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_char preprocess_buffer '"';
           string oneline_annot lexbuf )
# 1251 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 4 ->
# 489 "src/kernel_internals/parsing/logic_preprocess.mll"
         ( Buffer.add_string preprocess_buffer "//";
           oneline_annot_comment lexbuf )
# 1257 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 5 ->
let
# 491 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1263 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 491 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( Buffer.add_char preprocess_buffer c;
             oneline_annot lexbuf )
# 1268 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_oneline_annot_rec lexbuf __ocaml_lex_state

and oneline_annot_comment lexbuf =
    __ocaml_lex_oneline_annot_comment_rec lexbuf 128
and __ocaml_lex_oneline_annot_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 495 "src/kernel_internals/parsing/logic_preprocess.mll"
             (
       incr curr_line;
       Buffer.add_char preprocess_buffer '\n';
       Buffer.add_string preprocess_buffer annot_end_comment;
       Buffer.add_char preprocess_buffer '\n';
       main lexbuf )
# 1285 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
let
# 501 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1291 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 501 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( Buffer.add_char preprocess_buffer c;
             oneline_annot_comment lexbuf )
# 1296 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_oneline_annot_comment_rec lexbuf __ocaml_lex_state

and oneline_comment lexbuf =
    __ocaml_lex_oneline_comment_rec lexbuf 131
and __ocaml_lex_oneline_comment_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 507 "src/kernel_internals/parsing/logic_preprocess.mll"
      ( make_newline();
        Buffer.add_string output_buffer (lexeme lexbuf);
        main lexbuf)
# 1310 "src/kernel_internals/parsing/logic_preprocess.ml"

  | 1 ->
let
# 510 "src/kernel_internals/parsing/logic_preprocess.mll"
         c
# 1316 "src/kernel_internals/parsing/logic_preprocess.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 510 "src/kernel_internals/parsing/logic_preprocess.mll"
           ( Buffer.add_char output_buffer c;
             oneline_comment lexbuf)
# 1321 "src/kernel_internals/parsing/logic_preprocess.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_oneline_comment_rec lexbuf __ocaml_lex_state

;;

# 513 "src/kernel_internals/parsing/logic_preprocess.mll"
 
  let file suffix cpp filename =
    reset ();
    let debug = Kernel.Debug_category.exists (fun x -> x = "parser") in
    let inchan = open_in_bin filename in
    let lex = Lexing.from_channel inchan in
    let ppname =
      Extlib.temp_file_cleanup_at_exit ~debug
        (Filename.basename filename) ".pp"
    in
    let ppfile = open_out ppname in
    main lex;
    preprocess_annots suffix cpp ppfile;
    close_in inchan;
    close_out ppfile;
    ppname

# 1346 "src/kernel_internals/parsing/logic_preprocess.ml"
