# 49 "src/kernel_internals/parsing/clexer.mll"
 
open Cparser
module H = Hashtbl
module E = Errorloc

let currentLoc () = Errorloc.currentLoc ()

let one_line_ghost = ref false
let is_oneline_ghost () = !one_line_ghost
let enter_oneline_ghost () = one_line_ghost := true
let exit_oneline_ghost () = one_line_ghost := false

let ghost_code = ref false
let is_ghost_code () = !ghost_code
let enter_ghost_code () = ghost_code := true
let exit_ghost_code () = ghost_code := false

let addComment c = Cabshelper.Comments.add (currentLoc()) c

(* Some debugging support for line numbers *)
let dbgToken (t: token) =
  if false then begin
    let dprintf fmt = Kernel.debug fmt in
    (match t with
         IDENT n -> dprintf "IDENT(%s)\n" n
       | LBRACE l -> dprintf "LBRACE(%d)\n" (fst l).Lexing.pos_lnum
       | RBRACE l -> dprintf "RBRACE(%d)\n" (fst l).Lexing.pos_lnum
       | IF l -> dprintf "IF(%d)\n" (fst l).Lexing.pos_lnum
       | SWITCH l -> dprintf "SWITCH(%d)\n" (fst l).Lexing.pos_lnum
       | RETURN l -> dprintf "RETURN(%d)\n" (fst l).Lexing.pos_lnum
       | _ -> ()) ;
    t
  end else
    t

(*
** Keyword hashtable
*)
let lexicon = H.create 211
let init_lexicon _ =
  H.clear lexicon;
  Logic_env.reset_typenames ();
  Logic_env.builtin_types_as_typenames ();
  List.iter
    (fun (key, builder) -> H.add lexicon key builder)
    [ ("auto", fun loc -> AUTO loc);
      ("const", fun loc -> CONST loc);
      ("__const", fun loc -> CONST loc);
      ("__const__", fun loc -> CONST loc);
      ("static", fun loc -> STATIC loc);
      ("extern", fun loc -> EXTERN loc);
      ("long", fun loc -> LONG loc);
      ("short", fun loc -> SHORT loc);
      ("register", fun loc -> REGISTER loc);
      ("signed", fun loc -> SIGNED loc);
      ("__signed", fun loc -> SIGNED loc);
      ("unsigned", fun loc -> UNSIGNED loc);
      ("volatile", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);
      (* WW: see /usr/include/sys/cdefs.h for why __signed and __volatile
       * are accepted GCC-isms *)
      ("char", fun loc -> CHAR loc);
      ("_Bool", fun loc -> BOOL loc);
      ("int", fun loc -> INT loc);
      ("float", fun loc -> FLOAT loc);
      ("double", fun loc -> DOUBLE loc);
      ("void", fun loc -> VOID loc);
      ("enum", fun loc -> ENUM loc);
      ("struct", fun loc -> STRUCT loc);
      ("typedef", fun loc -> TYPEDEF loc);
      ("union", fun loc -> UNION loc);
      ("break", fun loc -> BREAK loc);
      ("continue", fun loc -> CONTINUE loc);
      ("goto", fun loc -> GOTO loc);
      ("return", fun loc -> dbgToken (RETURN loc));
      ("switch", fun loc -> dbgToken (SWITCH loc));
      ("case", fun loc -> CASE loc);
      ("default", fun loc -> DEFAULT loc);
      ("while", fun loc -> WHILE loc);
      ("do", fun loc -> DO loc);
      ("for", fun loc -> FOR loc);
      ("if", fun loc -> dbgToken (IF loc));
      ("else", fun _ -> ELSE);
      (*** Implementation specific keywords ***)
      ("__signed__", fun loc -> SIGNED loc);
      ("__inline__", fun loc -> INLINE loc);
      ("inline", fun loc -> INLINE loc);
      ("__inline", fun loc -> INLINE loc);
      ("_inline", fun loc ->
                      if !Cprint.msvcMode then
                        INLINE loc
                      else
                        IDENT ("_inline"));
      ("__attribute__", fun loc -> ATTRIBUTE loc);
      ("__attribute", fun loc -> ATTRIBUTE loc);
(*
      ("__attribute_used__", fun loc -> ATTRIBUTE_USED loc);
*)
      ("__blockattribute__", fun _ -> BLOCKATTRIBUTE);
      ("__blockattribute", fun _ -> BLOCKATTRIBUTE);
      ("__asm__", fun loc -> ASM loc);
      ("asm", fun loc -> ASM loc);
      ("__typeof__", fun loc -> TYPEOF loc);
      ("__typeof", fun loc -> TYPEOF loc);
      ("typeof", fun loc -> TYPEOF loc);
      ("__alignof", fun loc -> ALIGNOF loc);
      ("__alignof__", fun loc -> ALIGNOF loc);
      ("__volatile__", fun loc -> VOLATILE loc);
      ("__volatile", fun loc -> VOLATILE loc);

      ("__FUNCTION__", fun loc -> FUNCTION__ loc);
      ("__func__", fun loc -> FUNCTION__ loc); (* ISO 6.4.2.2 *)
      ("__PRETTY_FUNCTION__", fun loc -> PRETTY_FUNCTION__ loc);
      ("__label__", fun _ -> LABEL__);
      (*** weimer: GCC arcana ***)
      ("__restrict", fun loc -> RESTRICT loc);
      ("restrict", fun loc -> RESTRICT loc);
(*      ("__extension__", EXTENSION); *)
      (**** MS VC ***)
      ("__int64", fun _ -> INT64 (currentLoc ()));
      ("__int32", fun loc -> INT loc);
      ("_cdecl",  fun _ -> MSATTR ("_cdecl", currentLoc ()));
      ("__cdecl", fun _ -> MSATTR ("__cdecl", currentLoc ()));
      ("_stdcall", fun _ -> MSATTR ("_stdcall", currentLoc ()));
      ("__stdcall", fun _ -> MSATTR ("__stdcall", currentLoc ()));
      ("_fastcall", fun _ -> MSATTR ("_fastcall", currentLoc ()));
      ("__fastcall", fun _ -> MSATTR ("__fastcall", currentLoc ()));
      ("__w64", fun _ -> MSATTR("__w64", currentLoc ()));
      ("__declspec", fun loc -> DECLSPEC loc);
      ("__forceinline", fun loc -> INLINE loc); (* !! we turn forceinline
                                                 * into inline *)
      ("__try", fun loc -> TRY loc);
      ("__except", fun loc -> EXCEPT loc);
      ("__finally", fun loc -> FINALLY loc);
      (* weimer: some files produced by 'GCC -E' expect this type to be
       * defined *)
      ("__builtin_va_list",
       fun _ -> NAMED_TYPE "__builtin_va_list");
      ("__builtin_va_arg", fun loc -> BUILTIN_VA_ARG loc);
      ("__builtin_types_compatible_p", fun loc -> BUILTIN_TYPES_COMPAT loc);
      ("__builtin_offsetof", fun loc -> BUILTIN_OFFSETOF loc);
      (* On some versions of GCC __thread is a regular identifier *)
      ("__thread",
       (fun loc ->
          if Cil.theMachine.Cil.theMachine.Cil_types.__thread_is_keyword then
            THREAD loc
          else
            IDENT "__thread"));
    ]


let is_c_keyword s = Hashtbl.mem lexicon s

(* Mark an identifier as a type name. The old mapping is preserved and will
 * be reinstated when we exit this context *)
let add_type name =
   (* ignore (print_string ("adding type name " ^ name ^ "\n"));  *)
  H.add lexicon name (fun _ -> NAMED_TYPE name);
  Logic_env.add_typename name

let context : string list list ref = ref [ [] ]

let push_context _ = context := []::!context

let pop_context _ =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
		(context := sub;
		List.iter (fun name ->
                            (* Format.eprintf
                              "removing lexicon for %s@." name; *)
                             H.remove lexicon name;
                             Logic_env.remove_typename name
                          ) con)

(* Mark an identifier as a variable name. The old mapping is preserved and
 * will be reinstated when we exit this context  *)
let add_identifier name =
  match !context with
    [] -> Kernel.fatal "Empty context stack"
  | con::sub ->
      (context := (name::con)::sub;
       (*Format.eprintf "adding IDENT for %s@." name;*)
       H.add lexicon name (fun _ -> dbgToken (IDENT name));
       Logic_env.hide_typename name
      )


(*
** Useful primitives
*)
let scan_ident id =
  let here = currentLoc () in
  try (H.find lexicon id) here
  (* default to variable name, as opposed to type *)
  with Not_found -> dbgToken (IDENT id)


(*
** Buffer processor
*)


let init ~(filename: string) : Lexing.lexbuf =
  init_lexicon ();
  (* Inititialize the pointer in Errormsg *)
  Lexerhack.add_type := add_type;
  Lexerhack.push_context := push_context;
  Lexerhack.pop_context := pop_context;
  Lexerhack.add_identifier := add_identifier;
  E.startParsing filename


let finish () =
  E.finishParsing ();
  Logic_env.reset_typenames ()


(*** escape character management ***)
let scan_escape (char: char) : int64 =
  let result = match char with
    'n' -> '\n'
  | 'r' -> '\r'
  | 't' -> '\t'
  | 'b' -> '\b'
  | 'f' -> '\012'  (* ASCII code 12 *)
  | 'v' -> '\011'  (* ASCII code 11 *)
  | 'a' -> '\007'  (* ASCII code 7 *)
  | 'e' | 'E' -> '\027'  (* ASCII code 27. This is a GCC extension *)
  | '\'' -> '\''
  | '"'-> '"'     (* '"' *)
  | '?' -> '?'
  | '(' when not !Cprint.msvcMode -> '('
  | '{' when not !Cprint.msvcMode -> '{'
  | '[' when not !Cprint.msvcMode -> '['
  | '%' when not !Cprint.msvcMode -> '%'
  | '\\' -> '\\'
  | other ->
    E.parse_error "Unrecognized escape sequence: \\%c" other
  in
  Int64.of_int (Char.code result)

let scan_hex_escape str =
  let radix = Int64.of_int 16 in
  let the_value = ref Int64.zero in
  (* start at character 2 to skip the \x *)
  for i = 2 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 16 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let scan_oct_escape str =
  let radix = Int64.of_int 8 in
  let the_value = ref Int64.zero in
  (* start at character 1 to skip the \x *)
  for i = 1 to (String.length str) - 1 do
    let thisDigit = Cabshelper.valueOfDigit (String.get str i) in
    (* the_value := !the_value * 8 + thisDigit *)
    the_value := Int64.add (Int64.mul !the_value radix) thisDigit
  done;
  !the_value

let lex_hex_escape remainder lexbuf =
  let prefix = scan_hex_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_oct_escape remainder lexbuf =
  let prefix = scan_oct_escape (Lexing.lexeme lexbuf) in
  prefix :: remainder lexbuf

let lex_simple_escape remainder lexbuf =
  let lexchar = Lexing.lexeme_char lexbuf 1 in
  let prefix = scan_escape lexchar in
  prefix :: remainder lexbuf

let lex_unescaped remainder lexbuf =
  let prefix = Int64.of_int (Char.code (Lexing.lexeme_char lexbuf 0)) in
  prefix :: remainder lexbuf

let lex_comment remainder buffer lexbuf =
  let ch = Lexing.lexeme_char lexbuf 0 in
  if ch = '\n' then E.newline() ;
  (match buffer with None -> () | Some b -> Buffer.add_char b ch) ;
  remainder buffer lexbuf

let do_lex_comment ?(first_string="") remainder lexbuf =
  let buffer =
    if Kernel.PrintComments.get () then begin
      let b = Buffer.create 80 in
      Buffer.add_string b first_string;
      Some b
    end else None
  in remainder buffer lexbuf ;
  match buffer with
    | Some b -> addComment (Buffer.contents b)
    | None -> ()


(* ISO standard locale-specific function to convert a wide character
 * into a sequence of normal characters. Here we work on strings.
 * We convert L"Hi" to "H\000i\000"
  matth: this seems unused.
let wbtowc wstr =
  let len = String.length wstr in
  let dest = String.make (len * 2) '\000' in
  for i = 0 to len-1 do
    dest.[i*2] <- wstr.[i] ;
  done ;
  dest
*)

(* This function converst the "Hi" in L"Hi" to { L'H', L'i', L'\0' }
  matth: this seems unused.
let wstr_to_warray wstr =
  let len = String.length wstr in
  let res = ref "{ " in
  for i = 0 to len-1 do
    res := !res ^ (Printf.sprintf "L'%c', " wstr.[i])
  done ;
  res := !res ^ "}" ;
  !res
*)

(* Pragmas get explicit end-of-line tokens.
 * Elsewhere they are silently discarded as whitespace. *)
let pragmaLine = ref false

let annot_char = ref '@'

let () =
  Kernel.ReadAnnot.add_set_hook
    (fun _ x ->
      (* prevent the C lexer interpretation of comments *)
      annot_char := if x then '@' else '\000');
  Kernel.CustomAnnot.add_set_hook (fun _ s -> annot_char:=s.[0])

let annot_start_pos = ref Cabshelper.cabslu
let buf = Buffer.create 1024

let save_current_pos () =
  annot_start_pos := currentLoc ()

let make_annot ~one_line lexbuf s =
  let start = snd !annot_start_pos in
  let stop, token = Logic_lexer.annot (start, s) in
  lexbuf.Lexing.lex_curr_p <- stop; 
  (* The filename has already been normalized, so we must reuse it "as is". *)
  E.setCurrentFile ~normalize:false stop.Lexing.pos_fname;
  E.setCurrentLine stop.Lexing.pos_lnum;
  if one_line then E.newline ();
  match token with
    | Logic_ptree.Adecl d -> DECL d
    | Logic_ptree.Aspec -> SPEC (start,s)
        (* At this point, we only have identified a function spec. Complete
           parsing of the annotation will only occur in the cparser.mly rule.
         *)
    | Logic_ptree.Acode_annot (loc,a) -> CODE_ANNOT (a, loc)
    | Logic_ptree.Aloop_annot (loc,a) -> LOOP_ANNOT (a,loc)
    | Logic_ptree.Aattribute_annot (loc,a) -> ATTRIBUTE_ANNOT (a, loc)
    | Logic_ptree.Acustom(loc,id, a) -> CUSTOM_ANNOT(a, id, loc)


# 368 "src/kernel_internals/parsing/clexer.ml"
let __ocaml_lex_tables = {
  Lexing.lex_base = 
   "\000\000\184\255\185\255\091\000\192\000\192\255\193\255\194\255\
    \195\255\196\255\197\255\198\255\199\255\200\255\201\255\202\255\
    \003\000\067\000\070\000\035\000\094\000\090\000\072\000\073\000\
    \091\000\172\000\074\000\170\000\013\001\051\001\244\255\124\001\
    \246\255\211\001\248\255\001\000\250\255\062\001\189\000\232\255\
    \220\000\080\001\254\255\059\000\255\255\252\255\060\000\253\255\
    \249\255\005\000\042\002\129\002\216\002\047\003\134\003\221\003\
    \052\004\139\004\226\004\057\005\144\005\231\005\062\006\149\006\
    \236\006\067\007\154\007\241\007\072\008\159\008\246\008\077\009\
    \164\009\251\009\082\010\169\010\000\011\243\255\245\255\089\000\
    \077\011\105\011\115\011\095\000\040\001\050\001\127\011\149\011\
    \096\000\144\011\146\011\161\011\014\001\065\001\064\001\140\011\
    \102\000\085\001\141\011\086\001\186\011\165\000\242\255\247\011\
    \001\012\030\012\040\012\200\001\069\012\118\002\079\012\205\002\
    \140\012\178\012\175\000\178\000\252\000\053\001\108\012\248\012\
    \216\012\025\013\036\003\058\013\123\003\240\255\168\000\156\000\
    \126\000\183\000\065\012\183\000\197\001\008\001\148\011\241\255\
    \251\000\162\000\131\000\199\000\206\000\236\255\222\255\138\000\
    \183\000\213\000\003\000\237\255\216\255\235\255\214\255\215\255\
    \234\255\233\255\231\255\206\255\230\255\207\255\229\255\228\255\
    \221\255\015\001\227\255\220\255\016\001\226\255\223\255\102\013\
    \189\013\020\014\107\014\194\014\031\001\255\255\081\001\253\255\
    \254\255\032\001\255\255\082\001\254\255\255\255\052\015\249\255\
    \250\255\043\001\252\255\253\255\074\002\255\255\251\255\057\015\
    \249\255\229\000\242\000\067\015\253\003\255\255\041\001\199\001\
    \052\001\151\011\253\255\068\001\048\001\021\001\090\001\033\001\
    \045\001\252\255\050\001\054\001\055\001\086\001\084\004\116\015\
    \073\001\080\001\148\001\115\001\102\001\103\001\146\001\168\001\
    \170\001\204\001\243\001\175\001\160\001\193\001\230\001\231\001\
    \242\001\251\255\250\001\035\002\002\002\250\001\245\001\034\002\
    \018\002\030\002\025\002\028\002\060\002\082\002\077\002\079\002\
    \081\002\090\002\097\002\078\002\119\002\122\002\106\002\118\002\
    \114\002\153\002\174\002\174\002\174\002\172\002\168\002\178\002\
    \185\002\206\002\188\002\193\002\200\002\215\002\241\002\004\003\
    \023\003\012\003\011\003\001\003\005\003\008\003\028\003\028\003\
    \042\003\044\003\035\003\086\003\078\003\092\003\091\003\090\003\
    \081\003\081\003\097\003\111\003\127\003\130\003\112\003\125\003\
    \147\003\149\003\111\013\251\255\092\015\171\004\255\255\252\255\
    \121\015\123\015\129\015\253\255\014\004\253\255\254\255\255\255\
    \015\004\254\255\255\255\139\010\250\255\251\255\164\012\255\255\
    \252\255\046\005\140\015\195\015\133\005\253\255\135\010\250\255\
    \251\255\218\015\255\255\252\255\220\005\233\015\045\016\051\006\
    \253\255\024\013\253\255\254\255\181\006\143\003\254\255\255\255\
    \162\012\253\255\165\003\255\255\164\003\148\003\155\003\254\255\
    \082\004\253\255\254\255\163\003\171\003\168\003\168\003\255\255\
    \102\004\252\255\253\255\254\255\009\004\255\255\169\004\254\255\
    \209\003\203\003\200\003\201\003\255\255\103\004\254\255\255\255\
    ";
  Lexing.lex_backtrk = 
   "\255\255\255\255\255\255\069\000\069\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \036\000\038\000\037\000\052\000\050\000\051\000\046\000\044\000\
    \043\000\042\000\047\000\064\000\017\000\017\000\255\255\069\000\
    \255\255\069\000\255\255\071\000\255\255\004\000\045\000\255\255\
    \002\000\000\000\255\255\001\000\255\255\255\255\003\000\255\255\
    \255\255\255\255\069\000\069\000\069\000\069\000\069\000\069\000\
    \008\000\069\000\069\000\069\000\069\000\066\000\069\000\069\000\
    \069\000\069\000\067\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\068\000\255\255\255\255\255\255\
    \255\255\255\255\013\000\255\255\017\000\017\000\017\000\016\000\
    \255\255\016\000\016\000\016\000\016\000\016\000\016\000\016\000\
    \255\255\017\000\017\000\017\000\017\000\255\255\255\255\255\255\
    \013\000\255\255\013\000\255\255\013\000\255\255\013\000\255\255\
    \255\255\015\000\255\255\015\000\015\000\015\000\255\255\255\255\
    \255\255\013\000\255\255\013\000\255\255\255\255\015\000\015\000\
    \015\000\255\255\014\000\255\255\014\000\014\000\014\000\255\255\
    \014\000\014\000\014\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\030\000\255\255\255\255\031\000\255\255\255\255\069\000\
    \069\000\069\000\069\000\065\000\001\000\255\255\255\255\255\255\
    \255\255\002\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\006\000\255\255\255\255\001\000\255\255\255\255\255\255\
    \255\255\006\000\006\000\002\000\001\000\255\255\255\255\002\000\
    \002\000\002\000\255\255\002\000\002\000\002\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\005\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\004\000\001\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\005\000\255\255\
    \255\255\002\000\003\000\001\000\002\000\255\255\255\255\255\255\
    \255\255\005\000\255\255\255\255\002\000\003\000\001\000\002\000\
    \255\255\255\255\255\255\255\255\000\000\255\255\255\255\255\255\
    \255\255\255\255\002\000\255\255\255\255\255\255\255\255\255\255\
    \003\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\003\000\255\255\002\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    ";
  Lexing.lex_default = 
   "\001\000\000\000\000\000\255\255\255\255\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \000\000\255\255\000\000\255\255\000\000\255\255\255\255\000\000\
    \045\000\042\000\000\000\255\255\000\000\000\000\255\255\000\000\
    \000\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\000\000\
    \255\255\255\255\255\255\255\255\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\255\255\000\000\000\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\000\000\175\000\000\000\
    \000\000\255\255\000\000\180\000\000\000\000\000\183\000\000\000\
    \000\000\255\255\000\000\000\000\255\255\000\000\000\000\192\000\
    \000\000\255\255\255\255\255\255\255\255\000\000\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\043\001\000\000\048\001\255\255\000\000\000\000\
    \048\001\048\001\048\001\000\000\053\001\000\000\000\000\000\000\
    \057\001\000\000\000\000\060\001\000\000\000\000\064\001\000\000\
    \000\000\255\255\255\255\255\255\255\255\000\000\071\001\000\000\
    \000\000\075\001\000\000\000\000\255\255\255\255\255\255\255\255\
    \000\000\082\001\000\000\000\000\255\255\086\001\000\000\000\000\
    \089\001\000\000\255\255\000\000\255\255\255\255\255\255\000\000\
    \255\255\000\000\000\000\255\255\255\255\255\255\255\255\000\000\
    \105\001\000\000\000\000\000\000\255\255\000\000\255\255\000\000\
    \255\255\255\255\255\255\255\255\000\000\118\001\000\000\000\000\
    ";
  Lexing.lex_trans = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\037\000\036\000\048\000\037\000\037\000\049\000\048\000\
    \000\000\000\000\049\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\037\000\000\000\000\000\000\000\000\000\000\000\
    \037\000\026\000\030\000\034\000\147\000\022\000\020\000\032\000\
    \008\000\007\000\023\000\025\000\005\000\024\000\027\000\038\000\
    \029\000\028\000\028\000\028\000\028\000\028\000\028\000\028\000\
    \028\000\028\000\014\000\006\000\018\000\016\000\017\000\015\000\
    \166\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\031\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\010\000\035\000\009\000\019\000\033\000\
    \159\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\004\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\012\000\021\000\011\000\013\000\003\000\
    \163\000\164\000\161\000\160\000\157\000\154\000\153\000\142\000\
    \151\000\130\000\130\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\101\000\096\000\156\000\
    \152\000\150\000\092\000\158\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\044\000\047\000\
    \044\000\047\000\003\000\143\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\155\000\148\000\
    \140\000\097\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\003\000\129\000\255\255\041\000\
    \125\000\149\000\125\000\125\000\040\000\139\000\135\000\135\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\039\000\135\000\141\000\125\000\127\000\144\000\
    \002\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\114\000\046\000\125\000\128\000\003\000\
    \145\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\167\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\082\000\082\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\037\000\
    \126\000\146\000\037\000\037\000\162\000\165\000\173\000\178\000\
    \135\000\125\000\081\000\081\000\136\000\190\000\082\000\210\000\
    \037\000\085\000\255\255\207\000\181\000\135\000\037\000\206\000\
    \082\000\082\000\084\000\087\000\087\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\028\000\028\000\081\000\082\000\082\000\
    \135\000\125\000\081\000\081\000\099\000\079\000\083\000\081\000\
    \081\000\086\000\255\255\177\000\202\000\135\000\098\000\085\000\
    \203\000\202\000\084\000\082\000\082\000\081\000\081\000\097\000\
    \084\000\202\000\125\000\080\000\092\000\081\000\202\000\208\000\
    \043\000\083\000\209\000\211\000\100\000\079\000\092\000\081\000\
    \081\000\202\000\081\000\081\000\083\000\212\000\077\000\086\000\
    \003\000\126\000\097\000\078\000\213\000\081\000\081\000\097\000\
    \084\000\202\000\125\000\080\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\092\000\214\000\
    \036\001\202\000\081\000\081\000\023\001\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\016\001\
    \002\001\249\000\248\000\003\000\255\255\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \106\000\106\000\244\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\239\000\236\000\235\000\
    \234\000\137\000\229\000\204\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\051\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\131\000\230\000\
    \198\000\138\000\050\000\205\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\231\000\
    \255\255\176\000\181\000\188\000\228\000\232\000\188\000\188\000\
    \227\000\233\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\188\000\233\000\233\000\237\000\
    \238\000\233\000\188\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\240\000\241\000\242\000\
    \243\000\003\000\233\000\059\000\003\000\003\000\003\000\057\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\058\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\108\000\108\000\
    \108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\
    \245\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\246\000\247\000\233\000\233\000\250\000\
    \251\000\252\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\253\000\254\000\255\000\000\001\
    \003\000\001\001\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\052\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\110\000\110\000\110\000\
    \110\000\110\000\110\000\110\000\110\000\110\000\110\000\233\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\001\004\001\005\001\006\001\007\001\008\001\
    \009\001\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\010\001\011\001\012\001\013\001\003\000\
    \014\001\053\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\015\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\233\000\017\001\018\001\019\001\020\001\021\001\022\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\233\000\024\001\025\001\026\001\003\000\027\001\
    \003\000\003\000\003\000\003\000\003\000\003\000\054\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\123\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\028\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \029\001\030\001\031\001\032\001\033\001\034\001\035\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\233\000\037\001\038\001\039\001\003\000\040\001\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\055\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\041\001\233\000\092\001\093\001\196\000\094\001\
    \095\001\196\000\196\000\100\001\087\001\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\196\000\
    \055\001\058\001\101\001\102\001\103\001\196\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \109\001\113\001\114\001\115\001\003\000\116\001\056\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\098\001\097\001\215\000\000\000\098\001\
    \215\000\215\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\215\000\000\000\
    \106\001\119\001\098\001\000\000\215\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\255\255\
    \108\001\000\000\098\001\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\111\001\000\000\045\001\000\000\111\001\045\001\
    \045\001\099\001\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\045\001\000\000\000\000\
    \000\000\111\001\000\000\045\001\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\111\001\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\067\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\054\001\255\255\
    \112\001\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\062\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\068\001\107\001\255\255\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\060\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\069\001\069\001\069\001\
    \069\001\069\001\069\001\069\001\069\001\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\061\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\079\001\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\080\001\080\001\080\001\080\001\080\001\
    \080\001\080\001\080\001\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\063\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\084\001\000\000\
    \000\000\084\001\084\001\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\084\001\
    \000\000\000\000\000\000\000\000\000\000\084\001\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\064\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\065\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\066\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \003\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\068\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\069\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\003\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\070\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\071\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\072\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\073\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\074\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\063\001\074\001\000\000\
    \000\000\075\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\073\001\000\000\000\000\000\000\062\001\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \076\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\000\000\000\000\000\000\000\000\003\000\
    \000\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\112\000\000\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\113\000\113\000\113\000\072\001\
    \000\000\000\000\000\000\061\001\000\000\000\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\111\000\000\000\111\000\000\000\
    \000\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\
    \110\000\110\000\110\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\104\000\104\000\082\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\000\000\000\000\000\000\000\000\
    \103\000\102\000\082\000\082\000\000\000\000\000\082\000\102\000\
    \082\000\000\000\000\000\082\000\081\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\028\000\028\000\082\000\
    \000\000\081\000\081\000\000\000\097\000\081\000\000\000\081\000\
    \103\000\102\000\081\000\000\000\094\000\000\000\093\000\102\000\
    \000\000\090\000\097\000\000\000\081\000\000\000\081\000\092\000\
    \082\000\135\000\089\000\098\000\202\000\000\000\000\000\000\000\
    \000\000\081\000\081\000\000\000\097\000\081\000\092\000\081\000\
    \092\000\088\000\081\000\000\000\095\000\000\000\088\000\081\000\
    \136\000\091\000\097\000\203\000\000\000\000\000\081\000\092\000\
    \000\000\135\000\089\000\000\000\202\000\093\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\092\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\081\000\
    \000\000\000\000\109\000\000\000\109\000\000\000\097\000\108\000\
    \108\000\108\000\108\000\108\000\108\000\108\000\108\000\108\000\
    \108\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\105\000\102\000\
    \000\000\107\000\000\000\107\000\000\000\102\000\106\000\106\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \106\000\106\000\000\000\000\000\000\000\000\000\105\000\102\000\
    \000\000\000\000\000\000\000\000\000\000\102\000\102\000\000\000\
    \000\000\130\000\130\000\000\000\102\000\108\000\108\000\108\000\
    \108\000\108\000\108\000\108\000\108\000\108\000\108\000\110\000\
    \110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\
    \110\000\000\000\000\000\102\000\000\000\133\000\102\000\000\000\
    \000\000\102\000\000\000\000\000\102\000\102\000\132\000\124\000\
    \000\000\124\000\000\000\102\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\123\000\123\000\000\000\000\000\
    \000\000\000\000\131\000\102\000\091\001\134\000\000\000\000\000\
    \000\000\102\000\000\000\000\000\000\000\102\000\132\000\000\000\
    \000\000\000\000\000\000\102\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\000\000\065\001\065\001\065\001\065\001\
    \065\001\065\001\065\001\065\001\000\000\091\001\000\000\000\000\
    \119\000\000\000\113\000\113\000\113\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\000\000\119\000\119\000\119\000\
    \119\000\119\000\119\000\113\000\113\000\113\000\113\000\113\000\
    \113\000\000\000\000\000\000\000\066\001\000\000\116\000\000\000\
    \000\000\090\001\118\000\122\000\000\000\122\000\000\000\115\000\
    \121\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\000\000\113\000\113\000\113\000\113\000\113\000\
    \113\000\000\000\000\000\114\000\066\001\000\000\117\000\091\001\
    \000\000\084\001\118\000\000\000\084\001\084\001\000\000\115\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\084\001\000\000\000\000\000\000\000\000\000\000\
    \084\001\119\000\119\000\119\000\119\000\119\000\119\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \120\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\119\000\119\000\119\000\119\000\119\000\119\000\102\000\
    \000\000\000\000\000\000\000\000\000\000\102\000\000\000\000\000\
    \120\000\000\000\123\000\123\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\000\000\000\000\000\000\000\000\
    \045\001\046\001\000\000\045\001\045\001\000\000\000\000\102\000\
    \102\000\000\000\000\000\000\000\000\000\102\000\102\000\000\000\
    \000\000\045\001\003\000\000\000\000\000\000\000\000\000\045\001\
    \000\000\044\001\000\000\083\001\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \102\000\000\000\255\255\000\000\255\255\000\000\102\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\003\000\000\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \168\000\003\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \255\255\000\000\000\000\000\000\003\000\000\000\003\000\003\000\
    \003\000\003\000\169\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\255\255\
    \000\000\000\000\000\000\003\000\000\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\170\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\000\000\000\000\
    \000\000\000\000\003\000\000\000\003\000\003\000\003\000\003\000\
    \003\000\171\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\000\000\000\000\000\000\
    \000\000\003\000\000\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\188\000\189\000\000\000\
    \188\000\188\000\196\000\197\000\000\000\196\000\196\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\188\000\000\000\
    \000\000\000\000\000\000\196\000\188\000\000\000\184\000\000\000\
    \000\000\196\000\000\000\000\000\187\000\186\000\000\000\000\000\
    \000\000\000\000\000\000\185\000\000\000\255\255\000\000\000\000\
    \255\255\195\000\195\000\195\000\195\000\195\000\195\000\195\000\
    \195\000\195\000\195\000\195\000\195\000\195\000\195\000\195\000\
    \195\000\195\000\195\000\195\000\195\000\215\000\047\001\000\000\
    \215\000\215\000\255\255\000\000\255\255\255\255\000\000\255\255\
    \000\000\000\000\255\255\049\001\000\000\255\255\215\000\200\000\
    \000\000\000\000\000\000\000\000\215\000\000\000\000\000\000\000\
    \199\000\000\000\000\000\047\001\000\000\047\001\000\000\000\000\
    \000\000\000\000\000\000\051\001\000\000\194\000\000\000\000\000\
    \049\001\193\000\050\001\000\000\198\000\000\000\000\000\201\000\
    \050\001\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \199\000\000\000\000\000\225\000\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\000\000\000\000\
    \218\000\000\000\000\000\000\000\000\000\067\001\067\001\067\001\
    \067\001\067\001\067\001\000\000\000\000\221\000\000\000\000\000\
    \000\000\000\000\000\000\217\000\000\000\224\000\000\000\000\000\
    \216\000\000\000\000\000\222\000\000\000\000\000\219\000\223\000\
    \000\000\220\000\000\000\226\000\000\000\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\067\001\067\001\067\001\067\001\
    \067\001\067\001\076\001\076\001\076\001\076\001\076\001\076\001\
    \076\001\076\001\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\078\001\078\001\078\001\078\001\078\001\078\001\078\001\
    \078\001\078\001\078\001\000\000\067\001\067\001\067\001\067\001\
    \067\001\067\001\078\001\078\001\078\001\078\001\078\001\078\001\
    \000\000\000\000\077\001\000\000\255\255\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\078\001\078\001\078\001\078\001\078\001\078\001\
    \000\000\000\000\077\001\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\255\255\078\001\078\001\078\001\
    \078\001\078\001\078\001\078\001\078\001\078\001\078\001\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\078\001\078\001\
    \078\001\078\001\078\001\078\001\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\255\255\000\000\000\000\000\000\000\000\
    \000\000\255\255\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\078\001\078\001\
    \078\001\078\001\078\001\078\001\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\255\255\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000";
  Lexing.lex_check = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\000\000\000\000\035\000\000\000\000\000\035\000\049\000\
    \255\255\255\255\049\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\000\000\255\255\255\255\255\255\255\255\255\255\
    \000\000\000\000\000\000\000\000\146\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \016\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \019\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\003\000\
    \017\000\017\000\018\000\018\000\020\000\022\000\023\000\026\000\
    \024\000\079\000\079\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\083\000\088\000\021\000\
    \024\000\024\000\096\000\020\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\043\000\046\000\
    \043\000\046\000\003\000\026\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\003\000\003\000\
    \003\000\003\000\003\000\003\000\003\000\003\000\021\000\025\000\
    \027\000\101\000\027\000\027\000\027\000\027\000\027\000\027\000\
    \027\000\027\000\027\000\027\000\004\000\114\000\040\000\038\000\
    \127\000\025\000\128\000\129\000\038\000\131\000\137\000\138\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\038\000\139\000\140\000\126\000\115\000\143\000\
    \000\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\115\000\040\000\126\000\115\000\004\000\
    \144\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\004\000\004\000\004\000\004\000\004\000\
    \004\000\004\000\004\000\028\000\092\000\028\000\028\000\028\000\
    \028\000\028\000\028\000\028\000\028\000\028\000\028\000\037\000\
    \116\000\145\000\037\000\037\000\161\000\164\000\172\000\177\000\
    \136\000\116\000\028\000\092\000\133\000\185\000\084\000\193\000\
    \037\000\028\000\041\000\194\000\179\000\133\000\037\000\198\000\
    \085\000\029\000\028\000\029\000\029\000\029\000\029\000\029\000\
    \029\000\029\000\029\000\029\000\029\000\084\000\094\000\093\000\
    \136\000\116\000\028\000\092\000\084\000\029\000\028\000\085\000\
    \029\000\028\000\041\000\174\000\204\000\133\000\085\000\029\000\
    \200\000\205\000\028\000\097\000\099\000\094\000\093\000\085\000\
    \029\000\200\000\117\000\029\000\094\000\084\000\206\000\207\000\
    \041\000\084\000\208\000\210\000\084\000\029\000\093\000\085\000\
    \029\000\203\000\097\000\099\000\029\000\211\000\031\000\029\000\
    \031\000\117\000\099\000\031\000\212\000\094\000\093\000\085\000\
    \029\000\200\000\117\000\029\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\093\000\213\000\
    \216\000\203\000\097\000\099\000\217\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\218\000\
    \219\000\220\000\221\000\031\000\040\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\031\000\
    \031\000\031\000\031\000\031\000\031\000\031\000\031\000\033\000\
    \107\000\107\000\107\000\107\000\107\000\107\000\107\000\107\000\
    \107\000\107\000\222\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\223\000\224\000\225\000\
    \227\000\132\000\228\000\199\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\132\000\229\000\
    \199\000\132\000\033\000\199\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\033\000\033\000\
    \033\000\033\000\033\000\033\000\033\000\033\000\050\000\230\000\
    \041\000\174\000\179\000\188\000\226\000\231\000\188\000\188\000\
    \226\000\232\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\188\000\234\000\235\000\236\000\
    \237\000\238\000\188\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\239\000\240\000\241\000\
    \242\000\050\000\243\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\050\000\050\000\050\000\
    \050\000\050\000\050\000\050\000\050\000\051\000\109\000\109\000\
    \109\000\109\000\109\000\109\000\109\000\109\000\109\000\109\000\
    \244\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\245\000\246\000\247\000\248\000\249\000\
    \250\000\251\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\252\000\253\000\254\000\255\000\
    \051\000\000\001\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\051\000\051\000\051\000\051\000\
    \051\000\051\000\051\000\051\000\052\000\111\000\111\000\111\000\
    \111\000\111\000\111\000\111\000\111\000\111\000\111\000\001\001\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\002\001\003\001\004\001\005\001\006\001\007\001\
    \008\001\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\009\001\010\001\011\001\012\001\052\000\
    \013\001\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\052\000\052\000\052\000\052\000\052\000\
    \052\000\052\000\052\000\053\000\122\000\122\000\122\000\122\000\
    \122\000\122\000\122\000\122\000\122\000\122\000\014\001\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\015\001\016\001\017\001\018\001\019\001\020\001\021\001\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\022\001\023\001\024\001\025\001\053\000\026\001\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\053\000\053\000\053\000\053\000\053\000\053\000\
    \053\000\053\000\054\000\124\000\124\000\124\000\124\000\124\000\
    \124\000\124\000\124\000\124\000\124\000\027\001\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \028\001\029\001\030\001\031\001\032\001\033\001\034\001\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\035\001\036\001\037\001\038\001\054\000\039\001\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\054\000\054\000\054\000\054\000\054\000\054\000\054\000\
    \054\000\055\000\040\001\041\001\090\001\092\001\196\000\093\001\
    \094\001\196\000\196\000\099\001\085\001\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\196\000\
    \052\001\056\001\100\001\101\001\102\001\196\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \108\001\112\001\113\001\114\001\055\000\115\001\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \055\000\055\000\055\000\055\000\055\000\055\000\055\000\055\000\
    \056\000\255\255\255\255\096\001\096\001\214\000\255\255\096\001\
    \214\000\214\000\255\255\255\255\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\214\000\255\255\
    \104\001\117\001\096\001\255\255\214\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\085\001\
    \104\001\255\255\096\001\056\000\255\255\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\056\000\
    \056\000\056\000\056\000\056\000\056\000\056\000\056\000\057\000\
    \255\255\255\255\110\001\255\255\045\001\255\255\110\001\045\001\
    \045\001\096\001\255\255\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\045\001\255\255\255\255\
    \255\255\110\001\255\255\045\001\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\255\255\255\255\
    \255\255\110\001\057\000\255\255\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\057\000\057\000\
    \057\000\057\000\057\000\057\000\057\000\057\000\058\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\052\001\056\001\
    \110\001\255\255\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\255\255\255\255\255\255\
    \255\255\058\000\255\255\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\058\000\058\000\058\000\
    \058\000\058\000\058\000\058\000\058\000\059\000\065\001\065\001\
    \065\001\065\001\065\001\065\001\065\001\065\001\104\001\117\001\
    \255\255\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\255\255\255\255\255\255\255\255\
    \059\000\255\255\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\059\000\059\000\059\000\059\000\
    \059\000\059\000\059\000\059\000\060\000\068\001\068\001\068\001\
    \068\001\068\001\068\001\068\001\068\001\255\255\255\255\255\255\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\255\255\255\255\255\255\255\255\060\000\
    \255\255\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\060\000\060\000\060\000\060\000\060\000\
    \060\000\060\000\060\000\061\000\076\001\076\001\076\001\076\001\
    \076\001\076\001\076\001\076\001\255\255\255\255\255\255\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\255\255\255\255\255\255\255\255\061\000\255\255\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\061\000\061\000\061\000\061\000\061\000\061\000\
    \061\000\061\000\062\000\079\001\079\001\079\001\079\001\079\001\
    \079\001\079\001\079\001\255\255\255\255\255\255\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\255\255\255\255\255\255\255\255\062\000\255\255\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\062\000\062\000\062\000\062\000\062\000\062\000\062\000\
    \062\000\063\000\255\255\255\255\255\255\255\255\084\001\255\255\
    \255\255\084\001\084\001\255\255\255\255\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\084\001\
    \255\255\255\255\255\255\255\255\255\255\084\001\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \255\255\255\255\255\255\255\255\063\000\255\255\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \063\000\063\000\063\000\063\000\063\000\063\000\063\000\063\000\
    \064\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\255\255\
    \255\255\255\255\255\255\064\000\255\255\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\064\000\
    \064\000\064\000\064\000\064\000\064\000\064\000\064\000\065\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\255\255\255\255\
    \255\255\255\255\065\000\255\255\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\065\000\065\000\
    \065\000\065\000\065\000\065\000\065\000\065\000\066\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\255\255\255\255\255\255\
    \255\255\066\000\255\255\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\066\000\066\000\066\000\
    \066\000\066\000\066\000\066\000\066\000\067\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\255\255\255\255\255\255\255\255\
    \067\000\255\255\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\067\000\067\000\067\000\067\000\
    \067\000\067\000\067\000\067\000\068\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\255\255\255\255\255\255\255\255\068\000\
    \255\255\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\068\000\068\000\068\000\068\000\068\000\
    \068\000\068\000\068\000\069\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\255\255\255\255\255\255\255\255\069\000\255\255\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\069\000\069\000\069\000\069\000\069\000\069\000\
    \069\000\069\000\070\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\255\255\255\255\255\255\255\255\070\000\255\255\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\070\000\070\000\070\000\070\000\070\000\070\000\070\000\
    \070\000\071\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \255\255\255\255\255\255\255\255\071\000\255\255\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \071\000\071\000\071\000\071\000\071\000\071\000\071\000\071\000\
    \072\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\255\255\
    \255\255\255\255\255\255\072\000\255\255\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\072\000\
    \072\000\072\000\072\000\072\000\072\000\072\000\072\000\073\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\255\255\255\255\
    \255\255\255\255\073\000\255\255\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\073\000\073\000\
    \073\000\073\000\073\000\073\000\073\000\073\000\074\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\059\001\070\001\255\255\
    \255\255\074\000\255\255\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\074\000\074\000\074\000\
    \074\000\074\000\074\000\074\000\074\000\075\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\070\001\255\255\255\255\255\255\059\001\
    \255\255\255\255\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\255\255\255\255\255\255\255\255\
    \075\000\255\255\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\075\000\075\000\075\000\075\000\
    \075\000\075\000\075\000\075\000\076\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\255\255\255\255\255\255\255\255\076\000\
    \255\255\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\076\000\076\000\076\000\076\000\076\000\
    \076\000\076\000\076\000\080\000\255\255\080\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\080\000\080\000\080\000\070\001\
    \255\255\255\255\255\255\059\001\255\255\255\255\080\000\080\000\
    \080\000\080\000\080\000\080\000\081\000\255\255\081\000\255\255\
    \255\255\081\000\081\000\081\000\081\000\081\000\081\000\081\000\
    \081\000\081\000\081\000\082\000\082\000\082\000\082\000\082\000\
    \082\000\082\000\082\000\082\000\082\000\086\000\080\000\080\000\
    \080\000\080\000\080\000\080\000\255\255\255\255\255\255\255\255\
    \082\000\082\000\095\000\098\000\255\255\255\255\089\000\082\000\
    \090\000\255\255\255\255\087\000\086\000\087\000\087\000\087\000\
    \087\000\087\000\087\000\087\000\087\000\087\000\087\000\091\000\
    \255\255\095\000\098\000\255\255\086\000\089\000\255\255\090\000\
    \082\000\082\000\087\000\255\255\089\000\255\255\090\000\082\000\
    \255\255\087\000\098\000\255\255\086\000\255\255\091\000\090\000\
    \100\000\134\000\087\000\086\000\201\000\255\255\255\255\255\255\
    \255\255\095\000\098\000\255\255\086\000\089\000\091\000\090\000\
    \095\000\089\000\087\000\255\255\089\000\255\255\087\000\100\000\
    \134\000\087\000\098\000\201\000\255\255\255\255\091\000\090\000\
    \255\255\134\000\087\000\255\255\201\000\091\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\091\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\100\000\
    \255\255\255\255\103\000\255\255\103\000\255\255\100\000\103\000\
    \103\000\103\000\103\000\103\000\103\000\103\000\103\000\103\000\
    \103\000\104\000\104\000\104\000\104\000\104\000\104\000\104\000\
    \104\000\104\000\104\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\104\000\104\000\
    \255\255\105\000\255\255\105\000\255\255\104\000\105\000\105\000\
    \105\000\105\000\105\000\105\000\105\000\105\000\105\000\105\000\
    \106\000\106\000\106\000\106\000\106\000\106\000\106\000\106\000\
    \106\000\106\000\255\255\255\255\255\255\255\255\104\000\104\000\
    \255\255\255\255\255\255\255\255\255\255\104\000\106\000\255\255\
    \255\255\130\000\130\000\255\255\106\000\108\000\108\000\108\000\
    \108\000\108\000\108\000\108\000\108\000\108\000\108\000\110\000\
    \110\000\110\000\110\000\110\000\110\000\110\000\110\000\110\000\
    \110\000\255\255\255\255\108\000\255\255\130\000\106\000\255\255\
    \255\255\108\000\255\255\255\255\106\000\110\000\130\000\118\000\
    \255\255\118\000\255\255\110\000\118\000\118\000\118\000\118\000\
    \118\000\118\000\118\000\118\000\118\000\118\000\255\255\255\255\
    \255\255\255\255\130\000\108\000\088\001\130\000\255\255\255\255\
    \255\255\108\000\255\255\255\255\255\255\110\000\130\000\255\255\
    \255\255\255\255\255\255\110\000\112\000\112\000\112\000\112\000\
    \112\000\112\000\112\000\112\000\112\000\112\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\112\000\112\000\112\000\
    \112\000\112\000\112\000\255\255\062\001\062\001\062\001\062\001\
    \062\001\062\001\062\001\062\001\255\255\088\001\255\255\255\255\
    \113\000\255\255\113\000\113\000\113\000\113\000\113\000\113\000\
    \113\000\113\000\113\000\113\000\255\255\112\000\112\000\112\000\
    \112\000\112\000\112\000\113\000\113\000\113\000\113\000\113\000\
    \113\000\255\255\255\255\255\255\062\001\255\255\113\000\255\255\
    \255\255\088\001\113\000\120\000\255\255\120\000\255\255\113\000\
    \120\000\120\000\120\000\120\000\120\000\120\000\120\000\120\000\
    \120\000\120\000\255\255\113\000\113\000\113\000\113\000\113\000\
    \113\000\255\255\255\255\113\000\062\001\255\255\113\000\088\001\
    \255\255\081\001\113\000\255\255\081\001\081\001\255\255\113\000\
    \119\000\119\000\119\000\119\000\119\000\119\000\119\000\119\000\
    \119\000\119\000\081\001\255\255\255\255\255\255\255\255\255\255\
    \081\001\119\000\119\000\119\000\119\000\119\000\119\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \119\000\121\000\121\000\121\000\121\000\121\000\121\000\121\000\
    \121\000\121\000\121\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\119\000\119\000\119\000\119\000\119\000\119\000\121\000\
    \255\255\255\255\255\255\255\255\255\255\121\000\255\255\255\255\
    \119\000\255\255\123\000\123\000\123\000\123\000\123\000\123\000\
    \123\000\123\000\123\000\123\000\255\255\255\255\255\255\255\255\
    \042\001\042\001\255\255\042\001\042\001\255\255\255\255\121\000\
    \123\000\255\255\255\255\255\255\255\255\121\000\123\000\255\255\
    \255\255\042\001\167\000\255\255\255\255\255\255\255\255\042\001\
    \255\255\042\001\255\255\081\001\255\255\255\255\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \123\000\255\255\088\001\255\255\062\001\255\255\123\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\255\255\255\255\255\255\255\255\167\000\255\255\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\167\000\167\000\167\000\167\000\167\000\167\000\167\000\
    \167\000\168\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \081\001\255\255\255\255\255\255\168\000\255\255\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \168\000\168\000\168\000\168\000\168\000\168\000\168\000\168\000\
    \169\000\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\169\000\169\000\169\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\169\000\169\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\169\000\042\001\
    \255\255\255\255\255\255\169\000\255\255\169\000\169\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\169\000\169\000\
    \169\000\169\000\169\000\169\000\169\000\169\000\169\000\170\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\170\000\170\000\170\000\170\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\170\000\170\000\170\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\255\255\255\255\
    \255\255\255\255\170\000\255\255\170\000\170\000\170\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\170\000\170\000\
    \170\000\170\000\170\000\170\000\170\000\170\000\171\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\255\255\255\255\255\255\
    \255\255\171\000\255\255\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\171\000\171\000\171\000\
    \171\000\171\000\171\000\171\000\171\000\182\000\182\000\255\255\
    \182\000\182\000\191\000\191\000\255\255\191\000\191\000\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\182\000\255\255\
    \255\255\255\255\255\255\191\000\182\000\255\255\182\000\255\255\
    \255\255\191\000\255\255\255\255\182\000\182\000\255\255\255\255\
    \255\255\255\255\255\255\182\000\255\255\044\001\255\255\255\255\
    \044\001\191\000\191\000\191\000\191\000\191\000\191\000\191\000\
    \191\000\191\000\191\000\195\000\195\000\195\000\195\000\195\000\
    \195\000\195\000\195\000\195\000\195\000\215\000\044\001\255\255\
    \215\000\215\000\048\001\255\255\049\001\048\001\255\255\049\001\
    \255\255\255\255\050\001\044\001\255\255\050\001\215\000\195\000\
    \255\255\255\255\255\255\255\255\215\000\255\255\255\255\255\255\
    \195\000\255\255\255\255\048\001\255\255\049\001\255\255\255\255\
    \255\255\255\255\255\255\050\001\255\255\191\000\255\255\255\255\
    \048\001\191\000\049\001\255\255\195\000\255\255\255\255\195\000\
    \050\001\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \195\000\255\255\255\255\215\000\066\001\066\001\066\001\066\001\
    \066\001\066\001\066\001\066\001\066\001\066\001\255\255\255\255\
    \215\000\255\255\255\255\255\255\255\255\066\001\066\001\066\001\
    \066\001\066\001\066\001\255\255\255\255\215\000\255\255\255\255\
    \255\255\255\255\255\255\215\000\255\255\215\000\255\255\255\255\
    \215\000\255\255\255\255\215\000\255\255\255\255\215\000\215\000\
    \255\255\215\000\255\255\215\000\255\255\066\001\066\001\066\001\
    \066\001\066\001\066\001\067\001\067\001\067\001\067\001\067\001\
    \067\001\067\001\067\001\067\001\067\001\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\067\001\067\001\067\001\067\001\
    \067\001\067\001\073\001\073\001\073\001\073\001\073\001\073\001\
    \073\001\073\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\077\001\077\001\077\001\077\001\077\001\077\001\077\001\
    \077\001\077\001\077\001\255\255\067\001\067\001\067\001\067\001\
    \067\001\067\001\077\001\077\001\077\001\077\001\077\001\077\001\
    \255\255\255\255\073\001\255\255\182\000\255\255\255\255\255\255\
    \255\255\191\000\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\077\001\077\001\077\001\077\001\077\001\077\001\
    \255\255\255\255\073\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\044\001\078\001\078\001\078\001\
    \078\001\078\001\078\001\078\001\078\001\078\001\078\001\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\078\001\078\001\
    \078\001\078\001\078\001\078\001\255\255\255\255\255\255\255\255\
    \255\255\048\001\255\255\049\001\255\255\255\255\255\255\255\255\
    \255\255\050\001\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\078\001\078\001\
    \078\001\078\001\078\001\078\001\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\073\001\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255";
  Lexing.lex_base_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\002\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\004\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_backtrk_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_default_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    ";
  Lexing.lex_trans_code = 
   "\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\001\000\000\000\001\000\001\000\001\000\001\000\001\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\001\000\000\000\001\000\000\000\000\000\000\000\
    \001\000\000\000\001\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\000\
    \000\000\000\000\000\000";
  Lexing.lex_check_code = 
   "\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\214\000\255\255\215\000\214\000\214\000\215\000\215\000\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\214\000\255\255\215\000\255\255\255\255\255\255\
    \214\000\255\255\215\000\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\255\
    \255\255\255\255\255\255";
  Lexing.lex_code = 
   "\255\001\255\255\000\001\255";
}

let rec initial lexbuf =
    __ocaml_lex_initial_rec lexbuf 0
and __ocaml_lex_initial_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 473 "src/kernel_internals/parsing/clexer.mll"
      (
        let s = Lexing.lexeme lexbuf in
        let first_string =
          if String.length s > 2 then
            String.sub s 2 (String.length s - 2)
          else ""
        in
	do_lex_comment ~first_string comment lexbuf ;
        initial lexbuf
      )
# 1854 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 484 "src/kernel_internals/parsing/clexer.mll"
                        c
# 1860 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 485 "src/kernel_internals/parsing/clexer.mll"
    ( if c = !annot_char then begin
      try
        save_current_pos ();
	Buffer.clear buf;
	annot_first_token lexbuf
      with Parsing.Parse_error when Kernel.ContinueOnAnnotError.get () ->
        let source = Lexing.lexeme_start_p lexbuf in
        Kernel.debug ~source "skipping annotation";
	initial lexbuf
      end else
	begin
	  do_lex_comment ~first_string:(String.make 1 c) comment lexbuf ;
          initial lexbuf
	end
    )
# 1878 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 502 "src/kernel_internals/parsing/clexer.mll"
    ( 
      let s = Lexing.lexeme lexbuf in
      let first_string =
        if String.length s > 2 then
          String.sub s 2 (String.length s - 2)
        else ""
      in
      do_lex_comment ~first_string onelinecomment lexbuf ;
      E.newline();
      if is_oneline_ghost () then begin
        exit_oneline_ghost ();
        RGHOST
      end else begin
        initial lexbuf
      end
    )
# 1898 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
let
# 519 "src/kernel_internals/parsing/clexer.mll"
                    c
# 1904 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf (lexbuf.Lexing.lex_start_pos + 2) in
# 520 "src/kernel_internals/parsing/clexer.mll"
    ( if c = !annot_char then begin
      try
        save_current_pos ();
	Buffer.clear buf;
	annot_one_line lexbuf
      with Parsing.Parse_error when Kernel.ContinueOnAnnotError.get () ->
        let source = Lexing.lexeme_start_p lexbuf in
        Kernel.debug ~source "skipping annotation";
        initial lexbuf
      end else
	begin
	  do_lex_comment ~first_string:(String.make 1 c) onelinecomment lexbuf;
	  E.newline();
	  if is_oneline_ghost () then
	    begin
              exit_oneline_ghost ();
              RGHOST
	    end
	  else
	    begin
              initial lexbuf
	    end
	end
    )
# 1931 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 544 "src/kernel_internals/parsing/clexer.mll"
           (initial lexbuf)
# 1936 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 545 "src/kernel_internals/parsing/clexer.mll"
                                        ( E.newline ();
                                          if !pragmaLine then
                                            begin
                                              pragmaLine := false;
                                              PRAGMA_EOL
                                            end
                                          else if is_oneline_ghost () then
                                            begin
                                              exit_oneline_ghost ();
                                              RGHOST
                                            end
                                          else begin
                                            initial lexbuf
                                          end )
# 1954 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 559 "src/kernel_internals/parsing/clexer.mll"
                                        ( E.newline ();
                                          initial lexbuf
                                        )
# 1961 "src/kernel_internals/parsing/clexer.ml"

  | 7 ->
# 562 "src/kernel_internals/parsing/clexer.mll"
         ( hash lexbuf)
# 1966 "src/kernel_internals/parsing/clexer.ml"

  | 8 ->
# 563 "src/kernel_internals/parsing/clexer.mll"
                                   ( PRAGMA (currentLoc ()) )
# 1971 "src/kernel_internals/parsing/clexer.ml"

  | 9 ->
# 564 "src/kernel_internals/parsing/clexer.mll"
          ( CST_CHAR (chr lexbuf, currentLoc ()))
# 1976 "src/kernel_internals/parsing/clexer.ml"

  | 10 ->
# 565 "src/kernel_internals/parsing/clexer.mll"
          ( CST_WCHAR (chr lexbuf, currentLoc ()) )
# 1981 "src/kernel_internals/parsing/clexer.ml"

  | 11 ->
# 566 "src/kernel_internals/parsing/clexer.mll"
         (  
(* matth: BUG:  this could be either a regular string or a wide string.
 *  e.g. if it's the "world" in
 *     L"Hello, " "world"
 *  then it should be treated as wide even though there's no L immediately
 *  preceding it.  See test/small1/wchar5.c for a failure case. *)
                                          CST_STRING (str lexbuf, currentLoc()))
# 1992 "src/kernel_internals/parsing/clexer.ml"

  | 12 ->
# 573 "src/kernel_internals/parsing/clexer.mll"
           ( (* weimer: wchar_t string literal *)
                                          CST_WSTRING(str lexbuf, currentLoc()))
# 1998 "src/kernel_internals/parsing/clexer.ml"

  | 13 ->
# 575 "src/kernel_internals/parsing/clexer.mll"
             (CST_FLOAT (Lexing.lexeme lexbuf, currentLoc ()))
# 2003 "src/kernel_internals/parsing/clexer.ml"

  | 14 ->
# 576 "src/kernel_internals/parsing/clexer.mll"
                           ( (* GCC Extension for binary numbers *) 
                                          CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2009 "src/kernel_internals/parsing/clexer.ml"

  | 15 ->
# 578 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2014 "src/kernel_internals/parsing/clexer.ml"

  | 16 ->
# 579 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2019 "src/kernel_internals/parsing/clexer.ml"

  | 17 ->
# 580 "src/kernel_internals/parsing/clexer.mll"
            (CST_INT (Lexing.lexeme lexbuf, currentLoc ()))
# 2024 "src/kernel_internals/parsing/clexer.ml"

  | 18 ->
# 581 "src/kernel_internals/parsing/clexer.mll"
             (EOF)
# 2029 "src/kernel_internals/parsing/clexer.ml"

  | 19 ->
# 582 "src/kernel_internals/parsing/clexer.mll"
           (ELLIPSIS)
# 2034 "src/kernel_internals/parsing/clexer.ml"

  | 20 ->
# 583 "src/kernel_internals/parsing/clexer.mll"
          (PLUS_EQ)
# 2039 "src/kernel_internals/parsing/clexer.ml"

  | 21 ->
# 584 "src/kernel_internals/parsing/clexer.mll"
          (MINUS_EQ)
# 2044 "src/kernel_internals/parsing/clexer.ml"

  | 22 ->
# 585 "src/kernel_internals/parsing/clexer.mll"
          (STAR_EQ)
# 2049 "src/kernel_internals/parsing/clexer.ml"

  | 23 ->
# 586 "src/kernel_internals/parsing/clexer.mll"
          (SLASH_EQ)
# 2054 "src/kernel_internals/parsing/clexer.ml"

  | 24 ->
# 587 "src/kernel_internals/parsing/clexer.mll"
          (PERCENT_EQ)
# 2059 "src/kernel_internals/parsing/clexer.ml"

  | 25 ->
# 588 "src/kernel_internals/parsing/clexer.mll"
          (PIPE_EQ)
# 2064 "src/kernel_internals/parsing/clexer.ml"

  | 26 ->
# 589 "src/kernel_internals/parsing/clexer.mll"
          (AND_EQ)
# 2069 "src/kernel_internals/parsing/clexer.ml"

  | 27 ->
# 590 "src/kernel_internals/parsing/clexer.mll"
          (CIRC_EQ)
# 2074 "src/kernel_internals/parsing/clexer.ml"

  | 28 ->
# 591 "src/kernel_internals/parsing/clexer.mll"
           (INF_INF_EQ)
# 2079 "src/kernel_internals/parsing/clexer.ml"

  | 29 ->
# 592 "src/kernel_internals/parsing/clexer.mll"
           (SUP_SUP_EQ)
# 2084 "src/kernel_internals/parsing/clexer.ml"

  | 30 ->
# 593 "src/kernel_internals/parsing/clexer.mll"
          (INF_INF)
# 2089 "src/kernel_internals/parsing/clexer.ml"

  | 31 ->
# 594 "src/kernel_internals/parsing/clexer.mll"
          (SUP_SUP)
# 2094 "src/kernel_internals/parsing/clexer.ml"

  | 32 ->
# 595 "src/kernel_internals/parsing/clexer.mll"
           (EQ_EQ)
# 2099 "src/kernel_internals/parsing/clexer.ml"

  | 33 ->
# 596 "src/kernel_internals/parsing/clexer.mll"
           (EXCLAM_EQ)
# 2104 "src/kernel_internals/parsing/clexer.ml"

  | 34 ->
# 597 "src/kernel_internals/parsing/clexer.mll"
          (INF_EQ)
# 2109 "src/kernel_internals/parsing/clexer.ml"

  | 35 ->
# 598 "src/kernel_internals/parsing/clexer.mll"
          (SUP_EQ)
# 2114 "src/kernel_internals/parsing/clexer.ml"

  | 36 ->
# 599 "src/kernel_internals/parsing/clexer.mll"
         (EQ)
# 2119 "src/kernel_internals/parsing/clexer.ml"

  | 37 ->
# 600 "src/kernel_internals/parsing/clexer.mll"
         (INF)
# 2124 "src/kernel_internals/parsing/clexer.ml"

  | 38 ->
# 601 "src/kernel_internals/parsing/clexer.mll"
         (SUP)
# 2129 "src/kernel_internals/parsing/clexer.ml"

  | 39 ->
# 602 "src/kernel_internals/parsing/clexer.mll"
          (PLUS_PLUS (currentLoc ()))
# 2134 "src/kernel_internals/parsing/clexer.ml"

  | 40 ->
# 603 "src/kernel_internals/parsing/clexer.mll"
          (MINUS_MINUS (currentLoc ()))
# 2139 "src/kernel_internals/parsing/clexer.ml"

  | 41 ->
# 604 "src/kernel_internals/parsing/clexer.mll"
          (ARROW)
# 2144 "src/kernel_internals/parsing/clexer.ml"

  | 42 ->
# 605 "src/kernel_internals/parsing/clexer.mll"
         (PLUS (currentLoc ()))
# 2149 "src/kernel_internals/parsing/clexer.ml"

  | 43 ->
# 606 "src/kernel_internals/parsing/clexer.mll"
         (MINUS (currentLoc ()))
# 2154 "src/kernel_internals/parsing/clexer.ml"

  | 44 ->
# 608 "src/kernel_internals/parsing/clexer.mll"
                    ( if is_ghost_code () then might_end_ghost lexbuf
                      else
                        STAR (currentLoc ()))
# 2161 "src/kernel_internals/parsing/clexer.ml"

  | 45 ->
# 611 "src/kernel_internals/parsing/clexer.mll"
         (SLASH)
# 2166 "src/kernel_internals/parsing/clexer.ml"

  | 46 ->
# 612 "src/kernel_internals/parsing/clexer.mll"
         (PERCENT)
# 2171 "src/kernel_internals/parsing/clexer.ml"

  | 47 ->
# 613 "src/kernel_internals/parsing/clexer.mll"
         (EXCLAM (currentLoc ()))
# 2176 "src/kernel_internals/parsing/clexer.ml"

  | 48 ->
# 614 "src/kernel_internals/parsing/clexer.mll"
          (AND_AND (currentLoc ()))
# 2181 "src/kernel_internals/parsing/clexer.ml"

  | 49 ->
# 615 "src/kernel_internals/parsing/clexer.mll"
          (PIPE_PIPE)
# 2186 "src/kernel_internals/parsing/clexer.ml"

  | 50 ->
# 616 "src/kernel_internals/parsing/clexer.mll"
         (AND (currentLoc ()))
# 2191 "src/kernel_internals/parsing/clexer.ml"

  | 51 ->
# 617 "src/kernel_internals/parsing/clexer.mll"
         (PIPE)
# 2196 "src/kernel_internals/parsing/clexer.ml"

  | 52 ->
# 618 "src/kernel_internals/parsing/clexer.mll"
         (CIRC)
# 2201 "src/kernel_internals/parsing/clexer.ml"

  | 53 ->
# 619 "src/kernel_internals/parsing/clexer.mll"
         (QUEST)
# 2206 "src/kernel_internals/parsing/clexer.ml"

  | 54 ->
# 620 "src/kernel_internals/parsing/clexer.mll"
         (COLON)
# 2211 "src/kernel_internals/parsing/clexer.ml"

  | 55 ->
# 621 "src/kernel_internals/parsing/clexer.mll"
                (TILDE (currentLoc ()))
# 2216 "src/kernel_internals/parsing/clexer.ml"

  | 56 ->
# 623 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (LBRACE (currentLoc ())))
# 2221 "src/kernel_internals/parsing/clexer.ml"

  | 57 ->
# 624 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (RBRACE (currentLoc ())))
# 2226 "src/kernel_internals/parsing/clexer.ml"

  | 58 ->
# 625 "src/kernel_internals/parsing/clexer.mll"
          (LBRACKET)
# 2231 "src/kernel_internals/parsing/clexer.ml"

  | 59 ->
# 626 "src/kernel_internals/parsing/clexer.mll"
          (RBRACKET)
# 2236 "src/kernel_internals/parsing/clexer.ml"

  | 60 ->
# 627 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (LPAREN (currentLoc ())) )
# 2241 "src/kernel_internals/parsing/clexer.ml"

  | 61 ->
# 628 "src/kernel_internals/parsing/clexer.mll"
          (RPAREN)
# 2246 "src/kernel_internals/parsing/clexer.ml"

  | 62 ->
# 629 "src/kernel_internals/parsing/clexer.mll"
               (dbgToken (SEMICOLON (currentLoc ())) )
# 2251 "src/kernel_internals/parsing/clexer.ml"

  | 63 ->
# 630 "src/kernel_internals/parsing/clexer.mll"
          (COMMA)
# 2256 "src/kernel_internals/parsing/clexer.ml"

  | 64 ->
# 631 "src/kernel_internals/parsing/clexer.mll"
          (DOT)
# 2261 "src/kernel_internals/parsing/clexer.ml"

  | 65 ->
# 632 "src/kernel_internals/parsing/clexer.mll"
             (SIZEOF (currentLoc ()))
# 2266 "src/kernel_internals/parsing/clexer.ml"

  | 66 ->
# 633 "src/kernel_internals/parsing/clexer.mll"
                                        ( if !Cprint.msvcMode then
                                             MSASM (msasm lexbuf, currentLoc ())
                                          else (ASM (currentLoc ())) )
# 2273 "src/kernel_internals/parsing/clexer.ml"

  | 67 ->
# 638 "src/kernel_internals/parsing/clexer.mll"
                                        ( let _ = matchingpars 0 lexbuf in
                                          initial lexbuf
                                        )
# 2280 "src/kernel_internals/parsing/clexer.ml"

  | 68 ->
# 644 "src/kernel_internals/parsing/clexer.mll"
                                        (initial lexbuf )
# 2285 "src/kernel_internals/parsing/clexer.ml"

  | 69 ->
# 645 "src/kernel_internals/parsing/clexer.mll"
           (scan_ident (Lexing.lexeme lexbuf))
# 2290 "src/kernel_internals/parsing/clexer.ml"

  | 70 ->
# 647 "src/kernel_internals/parsing/clexer.mll"
  ( if is_oneline_ghost() then begin
      exit_oneline_ghost (); RGHOST
    end
    else EOF
  )
# 2299 "src/kernel_internals/parsing/clexer.ml"

  | 71 ->
# 652 "src/kernel_internals/parsing/clexer.mll"
       (E.parse_error "Invalid symbol")
# 2304 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_initial_rec lexbuf __ocaml_lex_state

and might_end_ghost lexbuf =
    __ocaml_lex_might_end_ghost_rec lexbuf 172
and __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 655 "src/kernel_internals/parsing/clexer.mll"
        ( exit_ghost_code(); RGHOST )
# 2316 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 656 "src/kernel_internals/parsing/clexer.mll"
       ( STAR (currentLoc()) )
# 2321 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_might_end_ghost_rec lexbuf __ocaml_lex_state

and comment buffer lexbuf =
    __ocaml_lex_comment_rec buffer lexbuf 174
and __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 659 "src/kernel_internals/parsing/clexer.mll"
                (  )
# 2333 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 660 "src/kernel_internals/parsing/clexer.mll"
                ( E.parse_error "Unterminated C comment" )
# 2338 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 661 "src/kernel_internals/parsing/clexer.mll"
                ( lex_comment comment buffer lexbuf )
# 2343 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_comment_rec buffer lexbuf __ocaml_lex_state

and onelinecomment buffer lexbuf =
    __ocaml_lex_onelinecomment_rec buffer lexbuf 179
and __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 664 "src/kernel_internals/parsing/clexer.mll"
                (  )
# 2355 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 665 "src/kernel_internals/parsing/clexer.mll"
                ( lex_comment onelinecomment buffer lexbuf )
# 2360 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_onelinecomment_rec buffer lexbuf __ocaml_lex_state

and matchingpars parsopen lexbuf =
    __ocaml_lex_matchingpars_rec parsopen lexbuf 182
and __ocaml_lex_matchingpars_rec parsopen lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 668 "src/kernel_internals/parsing/clexer.mll"
                ( E.newline (); matchingpars parsopen lexbuf )
# 2372 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 669 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars parsopen lexbuf )
# 2377 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 670 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars (parsopen + 1) lexbuf )
# 2382 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 671 "src/kernel_internals/parsing/clexer.mll"
                ( if parsopen > 1 then
                    matchingpars (parsopen - 1) lexbuf
                )
# 2389 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 674 "src/kernel_internals/parsing/clexer.mll"
         ( do_lex_comment comment lexbuf ;
                  matchingpars parsopen lexbuf )
# 2395 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 676 "src/kernel_internals/parsing/clexer.mll"
        ( let _ = str lexbuf in
                  matchingpars parsopen lexbuf )
# 2401 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 678 "src/kernel_internals/parsing/clexer.mll"
                ( matchingpars parsopen lexbuf )
# 2406 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_matchingpars_rec parsopen lexbuf __ocaml_lex_state

and hash lexbuf =
  lexbuf.Lexing.lex_mem <- Array.make 2 (-1) ;   __ocaml_lex_hash_rec lexbuf 191
and __ocaml_lex_hash_rec lexbuf __ocaml_lex_state =
  match Lexing.new_engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 682 "src/kernel_internals/parsing/clexer.mll"
        ( E.newline (); initial lexbuf)
# 2418 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 683 "src/kernel_internals/parsing/clexer.mll"
         ( hash lexbuf)
# 2423 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 684 "src/kernel_internals/parsing/clexer.mll"
         ( (* We are seeing a line number. This is the number for the
                   * next line *)
                 let s = Lexing.lexeme lexbuf in
                 let lineno = try
                   int_of_string s
                 with Failure ("int_of_string") ->
                   (* the int is too big. *)
                   Kernel.warning "Bad line number in preprocessed file: %s" s;
                   (-1)
                 in
                 E.setCurrentLine (lineno - 1);
                  (* A file name may follow *)
		  file lexbuf )
# 2440 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 697 "src/kernel_internals/parsing/clexer.mll"
                ( hash lexbuf )
# 2445 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
let
# 700 "src/kernel_internals/parsing/clexer.mll"
                                     pragmaName
# 2451 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme lexbuf lexbuf.Lexing.lex_mem.(0) lexbuf.Lexing.lex_curr_pos in
# 701 "src/kernel_internals/parsing/clexer.mll"
                ( let here = currentLoc () in
                  PRAGMA_LINE (pragmaName ^ pragma lexbuf, here)
                )
# 2457 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 704 "src/kernel_internals/parsing/clexer.mll"
                ( pragmaLine := true; PRAGMA (currentLoc ()) )
# 2462 "src/kernel_internals/parsing/clexer.ml"

  | 6 ->
# 705 "src/kernel_internals/parsing/clexer.mll"
            ( endline lexbuf)
# 2467 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_hash_rec lexbuf __ocaml_lex_state

and file lexbuf =
    __ocaml_lex_file_rec lexbuf 298
and __ocaml_lex_file_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 708 "src/kernel_internals/parsing/clexer.mll"
                      (E.newline (); initial lexbuf)
# 2479 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 709 "src/kernel_internals/parsing/clexer.mll"
          (file lexbuf)
# 2484 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 712 "src/kernel_internals/parsing/clexer.mll"
                                             (
                                 let n = Lexing.lexeme lexbuf in
				 let n1 = String.sub n 1 ((String.length n) - 4) in
                                 E.setCurrentWorkingDirectory n1;
                                 endline lexbuf )
# 2493 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 717 "src/kernel_internals/parsing/clexer.mll"
                                (  (* '"' *)
                                   let n = Lexing.lexeme lexbuf in
                                   let n1 = String.sub n 1
                                       ((String.length n) - 2) in
                                   E.setCurrentFile n1;
				 endline lexbuf)
# 2503 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 724 "src/kernel_internals/parsing/clexer.mll"
      (endline lexbuf)
# 2508 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_file_rec lexbuf __ocaml_lex_state

and endline lexbuf =
    __ocaml_lex_endline_rec lexbuf 308
and __ocaml_lex_endline_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 727 "src/kernel_internals/parsing/clexer.mll"
                ( E.newline (); initial lexbuf)
# 2520 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 728 "src/kernel_internals/parsing/clexer.mll"
                                ( EOF )
# 2525 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 729 "src/kernel_internals/parsing/clexer.mll"
      ( endline lexbuf)
# 2530 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_endline_rec lexbuf __ocaml_lex_state

and pragma lexbuf =
    __ocaml_lex_pragma_rec lexbuf 312
and __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 732 "src/kernel_internals/parsing/clexer.mll"
                        ( E.newline (); "" )
# 2542 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 733 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (pragma lexbuf) )
# 2548 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_pragma_rec lexbuf __ocaml_lex_state

and str lexbuf =
    __ocaml_lex_str_rec lexbuf 315
and __ocaml_lex_str_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 737 "src/kernel_internals/parsing/clexer.mll"
                        ([])
# 2560 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 738 "src/kernel_internals/parsing/clexer.mll"
             (lex_hex_escape str lexbuf)
# 2565 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 739 "src/kernel_internals/parsing/clexer.mll"
             (lex_oct_escape str lexbuf)
# 2570 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 740 "src/kernel_internals/parsing/clexer.mll"
          (lex_simple_escape str lexbuf)
# 2575 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 741 "src/kernel_internals/parsing/clexer.mll"
                        (E.parse_error "unterminated string" )
# 2580 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 742 "src/kernel_internals/parsing/clexer.mll"
     (lex_unescaped str lexbuf)
# 2585 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_str_rec lexbuf __ocaml_lex_state

and chr lexbuf =
    __ocaml_lex_chr_rec lexbuf 326
and __ocaml_lex_chr_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 745 "src/kernel_internals/parsing/clexer.mll"
              ([])
# 2597 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 746 "src/kernel_internals/parsing/clexer.mll"
             (lex_hex_escape chr lexbuf)
# 2602 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 747 "src/kernel_internals/parsing/clexer.mll"
             (lex_oct_escape chr lexbuf)
# 2607 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 748 "src/kernel_internals/parsing/clexer.mll"
          (lex_simple_escape chr lexbuf)
# 2612 "src/kernel_internals/parsing/clexer.ml"

  | 4 ->
# 749 "src/kernel_internals/parsing/clexer.mll"
                        ( E.parse_error "unterminated char" )
# 2617 "src/kernel_internals/parsing/clexer.ml"

  | 5 ->
# 750 "src/kernel_internals/parsing/clexer.mll"
     (lex_unescaped chr lexbuf)
# 2622 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_chr_rec lexbuf __ocaml_lex_state

and msasm lexbuf =
    __ocaml_lex_msasm_rec lexbuf 337
and __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 753 "src/kernel_internals/parsing/clexer.mll"
                        ( msasm lexbuf )
# 2634 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 754 "src/kernel_internals/parsing/clexer.mll"
                        ( msasminbrace lexbuf )
# 2639 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 755 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasmnobrace lexbuf) )
# 2645 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasm_rec lexbuf __ocaml_lex_state

and msasminbrace lexbuf =
    __ocaml_lex_msasminbrace_rec lexbuf 341
and __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 759 "src/kernel_internals/parsing/clexer.mll"
                        ( "" )
# 2657 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 760 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in
                          cur ^ (msasminbrace lexbuf) )
# 2663 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasminbrace_rec lexbuf __ocaml_lex_state

and msasmnobrace lexbuf =
    __ocaml_lex_msasmnobrace_rec lexbuf 344
and __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 763 "src/kernel_internals/parsing/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 1;
                          "" )
# 2677 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 766 "src/kernel_internals/parsing/clexer.mll"
                        ( lexbuf.Lexing.lex_curr_pos <-
                               lexbuf.Lexing.lex_curr_pos - 5;
                          "" )
# 2684 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 769 "src/kernel_internals/parsing/clexer.mll"
                        ( let cur = Lexing.lexeme lexbuf in

                          cur ^ (msasmnobrace lexbuf) )
# 2691 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_msasmnobrace_rec lexbuf __ocaml_lex_state

and annot_first_token lexbuf =
    __ocaml_lex_annot_first_token_rec lexbuf 352
and __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 774 "src/kernel_internals/parsing/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      Buffer.clear buf;
      enter_ghost_code ();
      LGHOST
    )
# 2708 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 780 "src/kernel_internals/parsing/clexer.mll"
                         c
# 2714 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 780 "src/kernel_internals/parsing/clexer.mll"
                           ( Buffer.add_char buf c; annot_first_token lexbuf )
# 2718 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 781 "src/kernel_internals/parsing/clexer.mll"
         ( E.newline(); Buffer.add_char buf '\n'; annot_first_token lexbuf )
# 2723 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
# 782 "src/kernel_internals/parsing/clexer.mll"
       ( annot_token lexbuf )
# 2728 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_first_token_rec lexbuf __ocaml_lex_state

and annot_token lexbuf =
    __ocaml_lex_annot_token_rec lexbuf 360
and __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 784 "src/kernel_internals/parsing/clexer.mll"
         ( let s = Buffer.contents buf in
           make_annot ~one_line:false lexbuf s )
# 2741 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
# 786 "src/kernel_internals/parsing/clexer.mll"
         ( E.parse_error "Unterminated annotation" )
# 2746 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 787 "src/kernel_internals/parsing/clexer.mll"
         (E.newline(); Buffer.add_char buf '\n'; annot_token lexbuf )
# 2751 "src/kernel_internals/parsing/clexer.ml"

  | 3 ->
let
# 788 "src/kernel_internals/parsing/clexer.mll"
         c
# 2757 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 788 "src/kernel_internals/parsing/clexer.mll"
           ( Buffer.add_char buf c; annot_token lexbuf )
# 2761 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_token_rec lexbuf __ocaml_lex_state

and annot_one_line lexbuf =
    __ocaml_lex_annot_one_line_rec lexbuf 366
and __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 791 "src/kernel_internals/parsing/clexer.mll"
            (
      if is_oneline_ghost () then E.parse_error "nested ghost code";
      enter_oneline_ghost (); LGHOST
    )
# 2776 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 795 "src/kernel_internals/parsing/clexer.mll"
                         c
# 2782 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 795 "src/kernel_internals/parsing/clexer.mll"
                           ( Buffer.add_char buf c; annot_one_line lexbuf )
# 2786 "src/kernel_internals/parsing/clexer.ml"

  | 2 ->
# 796 "src/kernel_internals/parsing/clexer.mll"
       ( annot_one_line_logic lexbuf )
# 2791 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_one_line_rec lexbuf __ocaml_lex_state

and annot_one_line_logic lexbuf =
    __ocaml_lex_annot_one_line_logic_rec lexbuf 373
and __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state =
  match Lexing.engine __ocaml_lex_tables __ocaml_lex_state lexbuf with
      | 0 ->
# 798 "src/kernel_internals/parsing/clexer.mll"
         ( make_annot ~one_line:true lexbuf (Buffer.contents buf) )
# 2803 "src/kernel_internals/parsing/clexer.ml"

  | 1 ->
let
# 799 "src/kernel_internals/parsing/clexer.mll"
         c
# 2809 "src/kernel_internals/parsing/clexer.ml"
= Lexing.sub_lexeme_char lexbuf lexbuf.Lexing.lex_start_pos in
# 799 "src/kernel_internals/parsing/clexer.mll"
           ( Buffer.add_char buf c; annot_one_line_logic lexbuf )
# 2813 "src/kernel_internals/parsing/clexer.ml"

  | __ocaml_lex_state -> lexbuf.Lexing.refill_buff lexbuf; 
      __ocaml_lex_annot_one_line_logic_rec lexbuf __ocaml_lex_state

;;

# 801 "src/kernel_internals/parsing/clexer.mll"
 

  (* Catch the exceptions raised by the lexer itself *)
  let initial lexbuf =
    try initial lexbuf
    with Failure _ -> raise Parsing.Parse_error 


# 2829 "src/kernel_internals/parsing/clexer.ml"
